from typing import Optional, Any, Union

from pint import Quantity, UndefinedUnitError, DimensionalityError
from pint.facets.plain import PlainQuantity

import enbios
from enbios.base.adapters_aggregators.aggregator import EnbiosAggregator
from enbios.generic.enbios2_logging import get_logger
from enbios.generic.tree.basic_tree import BasicTreeNode
from enbios.models.experiment_base_models import NodeOutput
from enbios.models.experiment_models import ScenarioResultNodeData, ResultValue


class SumAggregator(EnbiosAggregator):
    def __init__(self):
        self.logger = get_logger(__name__)

    def validate_config(self, config: Optional[dict[str, Any]]):
        pass

    def validate_node(self, node_name: str, node_config: Any):
        pass

    def aggregate_node_output(
        self,
        node: BasicTreeNode[ScenarioResultNodeData],
        scenario_name: Optional[str] = "",
    ) -> Optional[NodeOutput]:
        node_output: Optional[Union[Quantity, PlainQuantity]] = None
        for child in node.children:
            # if not child._data:
            #     raise ValueError(f"Node {child.name} has no data")
            if not child.data.output:
                continue
            node_output_unit = child.data.output.unit
            if node_output_unit is None:
                node_output = None
                self.logger.warning(f"No output unit of node '{child.name}'.")
                break
            output_mag: Optional[Quantity] = None
            try:
                output_mag = (
                    enbios.get_enbios_ureg().parse_expression(node_output_unit)
                    * child.data.output.magnitude
                )
                if not node_output:
                    node_output = output_mag
                else:
                    node_output += output_mag
            except UndefinedUnitError as err:
                self.logger.error(
                    f"Cannot parse output unit '{node_output_unit}' of node "
                    f"{child.name}. {err}. "
                    f"Consider the unit definition to 'enbios2/base/unit_registry.py'"
                )
                node_output = None
                break
            except DimensionalityError as err:
                set_base_unit = node_output.to_base_units().units if node_output else ""
                self.logger.warning(
                    f"Cannot aggregate output to parent: '{node.name}'. "
                    f"From earlier children the base unit is '{set_base_unit}'"
                    f"and from '{child.name}' it is '{output_mag.units}'."
                    f" {err}"
                )
                node_output = None
                break
        if node_output is not None:
            node_output = node_output.to_compact()
            result_node_output = NodeOutput(
                unit=str(node_output.units), magnitude=node_output.magnitude
            )
            return result_node_output
        else:
            # node.set_data(TechTreeNodeData())
            self.logger.warning(
                f"Scenario: '{scenario_name}': No output for node: '{node.name}' "
                f"(lvl: {node.level}). "
                f"Not calculating any upper nodes."
            )
            return None

    def aggregate_node_result(
        self, node: BasicTreeNode[ScenarioResultNodeData]
    ) -> dict[str, ResultValue]:
        result: dict[str, ResultValue] = {}
        for child in node.children:
            for key, value in child.data.results.items():
                ignore_short_multi_mag = False
                if result.get(key) is None:
                    result[key] = ResultValue(
                        magnitude=0, unit=value.unit, multi_magnitude=[]
                    )
                    ignore_short_multi_mag = True
                node_agg_result = result[key]
                if value.magnitude:
                    node_agg_result.magnitude += value.magnitude
                max_len = max(
                    len(node_agg_result.multi_magnitude), len(value.multi_magnitude)
                )
                if len(node_agg_result.multi_magnitude) < max_len:
                    node_agg_result.multi_magnitude.extend(
                        [0] * (max_len - len(node_agg_result.multi_magnitude))
                    )
                    if not ignore_short_multi_mag:
                        self.get_logger().warning(
                            f"Multi magnitude of node {node.name} is shorter than child {child.name}."
                        )
                if len(value.multi_magnitude) < max_len:
                    value.multi_magnitude.extend(
                        [0] * (max_len - len(value.multi_magnitude))
                    )
                    self.get_logger().warning(
                        f"Multi magnitude of child {child.name} is shorter than node {node.name}."
                    )

                result[key].multi_magnitude = [
                    a + b
                    for a, b in zip(
                        node_agg_result.multi_magnitude, value.multi_magnitude
                    )
                ]
        return result

    @staticmethod
    def node_indicator() -> str:
        return "sum"

    @staticmethod
    def name() -> str:
        return "sum-aggregator"

    @staticmethod
    def get_config_schemas() -> dict:
        return {}



import inspect
from pathlib import Path
from types import ModuleType
from typing import Union, Type

from enbios.base.adapters_aggregators.adapter import EnbiosAdapter
from enbios.base.adapters_aggregators.aggregator import EnbiosAggregator
from enbios.base.adapters_aggregators.builtin import BUILTIN_ADAPTERS, BUILTIN_AGGREGATORS
from enbios.generic.enbios2_logging import get_logger
from enbios.generic.util import load_module
from enbios.models.experiment_base_models import AdapterModel, AggregationModel

logger = get_logger(__name__)


def create_module_object(
    model_data: Union[AdapterModel, AggregationModel], base_class: Type
) -> Union[EnbiosAdapter, EnbiosAggregator]:
    if model_data.module_path:
        try:
            adapter_module: ModuleType = load_module(model_data.module_path)
        except Exception as err:
            raise ValueError(f"Could not load module '{model_data.module_path}' ({err})")
        # validator makes sure there is no other case
        for inspect_clazz in inspect.getmembers(adapter_module, inspect.isclass):
            # check if cl is subclass of EnbiosAdapter/EnbiosAggregation
            clazz = inspect_clazz[1]
            if any(base.__name__ == base_class.__name__ for base in clazz.__bases__):
                return clazz()
    else:
        if base_class.__name__ == EnbiosAdapter.__name__:
            clazz = BUILTIN_ADAPTERS.get(model_data.adapter_name)
            if not clazz:
                logger.error(
                    f"Could not find built-in adapter with name '{model_data.adapter_name}'. Candidates are: {BUILTIN_ADAPTERS.keys()}"
                )
            else:
                return clazz()
        else:
            assert base_class.__name__ == EnbiosAggregator.__name__
            clazz = BUILTIN_AGGREGATORS.get(model_data.aggregator_name)
            if not clazz:
                logger.error(
                    f"Could not find built-in aggregator with name '{model_data.aggregator_name}'. Candidates are: {BUILTIN_AGGREGATORS.keys()}"
                )
            else:
                return clazz()
    raise ValueError(
        f"'{Path(model_data.module_path).name}' has no class that inherits from '{base_class.__name__}'"
    )


def load_adapter(adapter_model: AdapterModel) -> EnbiosAdapter:
    return create_module_object(adapter_model, EnbiosAdapter)


def load_aggregator(aggregator_model: AggregationModel) -> EnbiosAggregator:
    return create_module_object(aggregator_model, EnbiosAggregator)



from csv import DictWriter

from bw2calc import LCA
from bw2data.backends import ActivityDataset


def solve_bioflows(lca: LCA) -> list[tuple[ActivityDataset, float]]:
    assert hasattr(lca, "characterized_inventory"), "Must do lcia first"
    # characterized_inventory:matrix biosphere (rows) x technosphere (cols), sum each row
    # solved_bioflows = lca.characterized_inventory.sum(1)
    # turn the result into a simple list
    # flat_solved_bioflows = solved_bioflows.flatten().tolist()[0]
    flat_solved_bioflows = lca.characterized_inventory.sum(1).flatten().tolist()[0]

    # collect all Activities from the database that are in the biosphere
    # biosphere_activities_ids = list(lca.dicts.biosphere.keys())
    # biosphere_activities = list(ActivityDataset.select().where(ActivityDataset.id.in_(biosphere_activities_ids))
    biosphere_activities = list(
        ActivityDataset.select().where(
            ActivityDataset.id.in_(list(lca.dicts.biosphere.keys()))
        )
    )
    # Sorting according to the index in lca.dicts.biosphere
    biosphere_activities.sort(key=lambda a: lca.dicts.biosphere.get(a.id))
    return list(zip(biosphere_activities, flat_solved_bioflows))


def solved_bioflow2csv(
    solved_bioflow: list[tuple[ActivityDataset, float]], file_path: str
):
    header = ["exchange name", "compartment", "amount", "unit"]
    with open(file_path, "w") as fout:
        writer = DictWriter(fout, fieldnames=header)
        writer.writeheader()
        for activity, output in solved_bioflow:
            data = activity.data
            writer.writerow(
                {
                    "exchange name": data["name"],
                    "compartment": ", ".join(data["categories"]),
                    "amount": output,
                    "unit": data["unit"],
                }
            )


# some_lca = LCA({some_activity: 1})
# some_lca.lci()
# some_lca.lcia()
# solved_bioflow = solve_bioflows(some_lca)
# solved_bioflow2csv(solved_bioflow, "some_activity_bioflows.csv")



from pathlib import Path

import os
import sys

# Get the absolute path of the entry-point script
entry_point_script = Path(os.path.abspath(sys.argv[0]))

PROJECT_PATH = entry_point_script.parent
BASE_DATA_PATH = PROJECT_PATH / "data"
BASE_TEST_DATA_PATH = Path(__file__).parent.parent / "test/data"



from abc import ABC, abstractmethod
from typing import Optional, Any

from enbios.generic.enbios2_logging import get_logger
from enbios.generic.tree.basic_tree import BasicTreeNode
from enbios.models.experiment_base_models import NodeOutput
from enbios.models.experiment_models import ScenarioResultNodeData


class EnbiosAggregator(ABC):
    @abstractmethod
    def validate_config(self, config: Optional[dict[str, Any]]):
        pass

    @abstractmethod
    def validate_node(self, node_name: str, node_config: Any):
        pass

    @abstractmethod
    def aggregate_node_output(
        self,
        node: BasicTreeNode[ScenarioResultNodeData],
        scenario_name: Optional[str] = "",
    ) -> Optional[NodeOutput]:
        pass

    @abstractmethod
    def aggregate_node_result(self, node: BasicTreeNode[ScenarioResultNodeData]):
        pass

    @staticmethod
    @abstractmethod
    def node_indicator() -> str:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @staticmethod
    @abstractmethod
    def get_config_schemas() -> dict:
        pass

    def get_logger(self):
        return get_logger(f"__name__ ({self.name})")



from enbios.base.adapters_aggregators.builtin.sum_aggregator import SumAggregator
from enbios.base.adapters_aggregators.builtin.simple_assignment_adapter import (
    SimpleAssignmentAdapter,
)
from enbios.bw2.brightway_experiment_adapter import BrightwayAdapter

BUILTIN_ADAPTERS = {
    SimpleAssignmentAdapter.name(): SimpleAssignmentAdapter,
    BrightwayAdapter.name(): BrightwayAdapter,
}

BUILTIN_AGGREGATORS = {SumAggregator.name(): SumAggregator}



import math
from logging import getLogger
from typing import Optional, Any, Union, Sequence, Callable

import bw2data
import bw2data as bd
from bw2calc.dictionary_manager import ReversibleRemappableDictionary
from bw2data.backends import Activity, ActivityDataset
from numpy import ndarray
from pint import Quantity, UndefinedUnitError
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from pydantic_core import core_schema, PydanticOmit

from enbios import get_enbios_ureg
from enbios.base.adapters_aggregators.adapter import EnbiosAdapter
from enbios.base.scenario import Scenario
from enbios.bw2.bw_models import (
    ExperimentMethodPrepData,
    BWAdapterConfig,
    BrightwayActivityConfig,
    BWMethodDefinition,
    BWActivityData,
    NonLinearMethodConfig,
)
from enbios.bw2.stacked_MultiLCA import StackedMultiLCA, BWCalculationSetup
from enbios.bw2.stacked_MultiLCA_regio import RegioStackedMultiLCA
from enbios.bw2.util import bw_unit_fix, get_activity
from enbios.generic.util import load_module, get_module_functions
from enbios.models.experiment_base_models import NodeOutput, AdapterModel
from enbios.models.experiment_models import (
    ResultValue,
)

logger = getLogger(__file__)

ureg = get_enbios_ureg()


def _bw_activity_search(activity_id: dict) -> Activity:
    """
    Search for the activity in the brightway project
    :param activity_id:
    :return: brightway activity
    """
    id_ = BrightwayActivityConfig(**activity_id)
    bw_activity: Optional[Activity] = None
    if id_.code:
        if id_.database:
            bw_activity = bd.Database(id_.database).get(id_.code)
        else:
            bw_activity = get_activity(id_.code)
    elif id_.name:
        filters = {}
        search_in_dbs = [id_.database] if id_.database else bd.databases
        for db in search_in_dbs:
            if id_.location:
                filters["location"] = id_.location
                search_results = bd.Database(db).search(id_.name, filter=filters)
            else:
                search_results = bd.Database(db).search(id_.name)
                # filter exact name
                search_results = list(
                    filter(lambda a: a["name"] == id_.name, search_results)
                )
            if id_.unit:
                search_results = list(
                    filter(lambda a: a["unit"] == id_.unit, search_results)
                )
            if len(search_results) == 1:
                bw_activity = search_results[0]
                break
            elif len(search_results) > 1:
                activities_str = "\n".join(
                    [f'{str(a)} - {a["code"]}' for a in search_results]
                )
                raise ValueError(
                    f"There are more than one activity with the same name, "
                    f"try including  "
                    f"the code of the activity you want to use:\n{activities_str}"
                )
    if not bw_activity:
        raise ValueError(f"No activity found for {activity_id}")
    return bw_activity


class BrightwayAdapter(EnbiosAdapter):
    @staticmethod
    def name() -> str:
        return "brightway-adapter"

    def validate_definition(self, definition: AdapterModel):
        pass

    def __init__(self):
        super(BrightwayAdapter, self).__init__()
        self.config: BWAdapterConfig = BWAdapterConfig(bw_project="")
        self.activityMap: dict[str, BWActivityData] = {}
        self.methods: dict[str, ExperimentMethodPrepData] = {}
        self.scenario_calc_setups: dict[str, BWCalculationSetup] = (
            {}
        )  # scenario_alias to BWCalculationSetup
        self.raw_results: dict[str, list[ndarray]] = {}  # scenario_alias to results
        self.lca_objects: dict[
            str, list[Union[StackedMultiLCA, RegioStackedMultiLCA]]
        ] = {}  # scenario_alias to lca objects
        self.all_regions_set: bool = (
            False  # as part of first run_scenario, go through set_node_regions
        )

    @staticmethod
    def node_indicator() -> str:
        return "bw"

    def assert_all_codes_unique(self):
        all_activities = list(ActivityDataset.select())
        assert len(all_activities) == len(
            set([a.code for a in all_activities])
        ), "It is recommended that all activities have unique codes"

    def validate_config(self, config: Optional[dict[str, Any]]):
        self.config = BWAdapterConfig(**config)
        if self.config.use_k_bw_distributions < 1:
            raise ValueError(
                f"config.use_k_bw_distributions must be greater than 0, "
                f"but is {self.config.use_k_bw_distributions}"
            )

        if self.config.bw_project not in bd.projects:
            raise ValueError(f"Project {self.config.bw_project} not found")
        else:
            bd.projects.set_current(self.config.bw_project)
        self.assert_all_codes_unique()

    def validate_methods(self, methods: Optional[dict[str, Any]]) -> list[str]:
        assert methods, "Methods must be defined for brightway adapter"
        # validation
        BWMethodDefinition.model_validate(methods)

        def validate_method(method_id: Sequence[str]) -> ExperimentMethodPrepData:
            # todo: should complain, if the same method is passed twice
            bw_method = bd.methods.get(method_id)
            if not bw_method:
                raise ValueError(f"Method with id: {method_id} does not exist")
            unit = bw_method.get("unit", "undefined method unit")
            return ExperimentMethodPrepData(id=tuple(method_id), bw_method_unit=unit)

        self.methods: dict[str, ExperimentMethodPrepData] = {
            name: validate_method(method) for name, method in methods.items()
        }
        return list(self.methods.keys())

    def validate_node_output(
        self,
        node_name: str,
        target_output: NodeOutput,
    ) -> float:
        """
        validate and convert to the bw-activity unit
        :param node_name:
        :param target_output:
        :return:
        """
        try:
            target_quantity: Quantity = (
                ureg.parse_expression(
                    bw_unit_fix(target_output.unit), case_sensitive=False
                )
                * target_output.magnitude
            )
            bw_activity_unit = self.activityMap[node_name].bw_activity["unit"]
            return target_quantity.to(bw_unit_fix(bw_activity_unit)).magnitude
        except UndefinedUnitError as err:
            logger.error(
                f"Cannot parse output unit '{target_output.unit}'- "
                f"of activity {node_name}. {err}. "
                f"Consider the unit definition to 'enbios2/base/unit_registry.py'"
            )
            raise UndefinedUnitError(f"Unit error, {err}; For activity: {node_name}")

    def validate_node(self, node_name: str, node_config: Any):
        assert isinstance(
            node_config, dict
        ), f"Activity id (type: dict) must be defined for activity {node_name}"
        # get the brightway activity
        bw_activity = _bw_activity_search(node_config)

        self.activityMap[node_name] = BWActivityData(
            bw_activity=bw_activity,
            default_output=NodeOutput(unit=bw_unit_fix(bw_activity["unit"]), magnitude=1),
        )
        if "default_output" in node_config:
            self.activityMap[node_name].default_output.magnitude = (
                self.validate_node_output(
                    node_name, NodeOutput(**node_config["default_output"])
                )
            )
        if self.config.simple_regionalization.run_regionalization:
            if "enb_location" in node_config:
                bw_activity["enb_location"] = node_config["enb_location"]
                bw_activity.save()

    def get_default_output_value(self, node_name: str) -> float:
        return self.activityMap[node_name].default_output.magnitude

    def get_node_output_unit(self, node_name: str) -> str:
        return bw_unit_fix(self.activityMap[node_name].bw_activity["unit"])

    def get_method_unit(self, method_name: str) -> str:
        return self.methods[method_name].bw_method_unit

    def prepare_scenario(self, scenario: Scenario):
        inventory: list[dict[Activity, float]] = []
        for act_alias, activity in self.activityMap.items():
            try:
                act_output = scenario.structural_nodes_outputs[act_alias]
                inventory.append({activity.bw_activity: act_output})
            except KeyError:
                # todo not sure if that ever happens..
                if not scenario.config.exclude_defaults:
                    raise Exception(
                        f"Activity {act_alias} not found in scenario {scenario.name}"
                    )

        methods = [m.id for m in self.methods.values()]
        calculation_setup = BWCalculationSetup(scenario.name, inventory, methods)
        calculation_setup.register()
        self.scenario_calc_setups[scenario.name] = calculation_setup
        if (
            self.config.simple_regionalization.run_regionalization
            and not self.all_regions_set
        ):
            # memorize nodes from the tree in order to not delete their location
            keep_locations_of_activities: list[str] = []
            if self.config.simple_regionalization.clear_all_other_node_regions:
                keep_locations_of_activities = [
                    a.bw_activity["code"] for a in self.activityMap.values()
                ]

                range_length = 1000
                activities_to_reset = list(
                    ActivityDataset.select().where(
                        ActivityDataset.code.not_in(keep_locations_of_activities)
                    )
                )

                for range_start in range(
                    math.ceil(len(activities_to_reset) / range_length)
                ):
                    # noinspection PyUnresolvedReferences
                    # noinspection PyProtectedMember
                    with ActivityDataset._meta.database.atomic():
                        for a in activities_to_reset[
                            range_start * range_length : (range_start + 1) * range_length
                        ]:
                            a.data["enb_location"] = None
                            a.save()

            # set additional specified locations. 'set_node_regions' field
            activity_codes: list[str] = list(
                self.config.simple_regionalization.set_node_regions.keys()
            )
            # this approach is much faster than individual updates
            # noinspection PyUnresolvedReferences
            # noinspection PyProtectedMember
            with ActivityDataset._meta.database.atomic():
                activities_to_reset = list(
                    ActivityDataset.select().where(
                        ActivityDataset.code.in_(activity_codes)
                    )
                )
                # validate all activities are present
                if len(activities_to_reset) != len(activity_codes):
                    missing = set(activity_codes) - set(
                        a.code for a in activities_to_reset
                    )
                    logger.warning(
                        f"Some activities specified in 'set_node_regions' are not found: {missing}"
                    )
                for a in activities_to_reset:
                    a.data["enb_location"] = tuple(
                        self.config.simple_regionalization.set_node_regions[a.code]
                    )
                    a.save()  # This updates each user in the database

    def run_scenario(self, scenario: Scenario) -> dict[str, dict[str, ResultValue]]:
        self.prepare_scenario(scenario)
        use_distributions = self.config.use_k_bw_distributions > 1
        raw_results: list[ndarray] = []
        self.lca_objects[scenario.name] = []
        run_regionalization = self.config.simple_regionalization.run_regionalization
        method_activity_func_maps: Optional[
            dict[tuple[str, ...], dict[int, Callable[[float], float]]]
        ] = None
        if self.config.nonlinear_characterization:
            method_activity_func_maps = self.prepare_nonlinear_methods()
        for i in range(self.config.use_k_bw_distributions):
            if self.config.use_k_bw_distributions > 1:
                self.get_logger().info(
                    f"Brightway adapter: Run distribution {i + 1}/{self.config.use_k_bw_distributions}"
                )
            if run_regionalization:
                _lca = RegioStackedMultiLCA(
                    self.scenario_calc_setups[scenario.name],
                    self.config.simple_regionalization.select_regions,
                    use_distributions=use_distributions,
                    method_activity_func_maps=method_activity_func_maps,
                )
                raw_results.append(_lca.results)
            else:
                _lca = StackedMultiLCA(
                    self.scenario_calc_setups[scenario.name],
                    use_distributions,
                    method_activity_func_maps=method_activity_func_maps,
                )
                raw_results.append(_lca.results)
            if self.config.store_lca_object:
                self.lca_objects[scenario.name].append(_lca)

        if self.config.store_raw_results:
            self.raw_results[scenario.name] = raw_results

        return self._assign_results2nodes(
            raw_results, scenario, use_distributions, run_regionalization
        )

    def _assign_results2nodes(
        self,
        raw_results: list[ndarray],
        scenario: Scenario,
        use_distributions: bool,
        has_regionalization: bool,
    ):
        result_data: dict[str, Any] = {}
        for act_idx, act_alias in enumerate(self.activityMap.keys()):
            if act_alias not in scenario.structural_nodes_outputs:
                if not scenario.config.exclude_defaults:
                    raise ValueError(
                        f"Activity {act_alias} not found in scenario {scenario.name}"
                    )
                continue
            result_data[act_alias] = {}
            result_field = "multi_magnitude" if use_distributions else "magnitude"
            for m_idx, method in enumerate(self.methods.items()):
                method_name, method_data = method
                if has_regionalization:
                    for region_idx, region in enumerate(
                        self.config.simple_regionalization.select_regions
                    ):
                        method_result = ResultValue(unit=method_data.bw_method_unit)
                        method_res_values = [
                            res[act_idx, m_idx, region_idx] for res in raw_results
                        ]
                        setattr(
                            method_result,
                            result_field,
                            (
                                method_res_values
                                if use_distributions
                                else method_res_values[0]
                            ),
                        )
                        result_data[act_alias][f"{method_name}.{region}"] = method_result
                else:
                    method_result = ResultValue(unit=method_data.bw_method_unit)
                    method_res_values = [res[act_idx, m_idx] for res in raw_results]
                    setattr(
                        method_result,
                        result_field,
                        method_res_values if use_distributions else method_res_values[0],
                    )
                    result_data[act_alias][method_name] = method_result
        return result_data

    def prepare_nonlinear_methods(self):
        # create dummy LCA to get the biosphere_matrix and prep_lca.dicts.biosphere
        # prep_lca = self.prepare_lca_for_nonlinear(calc_setup)
        config = self.config.nonlinear_characterization
        method_activity2func_maps = {}
        for method_name, method_config in config.methods.items():
            method_activity2func_map = self.prepare_nonlinear_method(
                method_name, method_config
            )
            bw_method_id = self.methods[method_name].id
            method_activity2func_maps[bw_method_id] = method_activity2func_map
        return method_activity2func_maps

    def prepare_nonlinear_method(
        self, method_name: str, method_config: NonLinearMethodConfig
    ) -> dict[int, Callable[[float], float]]:
        result_func_map: dict[int, Callable[[float], float]] = (
            {}
        )  # [None] * prep_lca.biosphere_matrix.shape[0]
        if method_name not in self.methods:
            raise ValueError(
                f"Unknown method '{method_name}' specified for nonlinear methods"
            )
        # bw method id (tuple[str,...])
        bw_method_id: tuple[str, ...] = self.methods[method_name].id
        if method_config.module_path_function_name:
            module_path, func_name = method_config.module_path_function_name
            func: Callable = get_module_functions(load_module(module_path)).get(func_name)
            if not func:
                raise ValueError(
                    f"Could not find function: '{func_name} in module: {module_path}, which is defined for"
                    f"brightway-adapter non-linear method definition for method: {method_name}"
                )
            method_config.functions = func()
        # key: (database,code) -> id
        biosphere_keys2ids = self.activities_keys_id_map(
            list(method_config.functions.keys())
        )
        for key, id_ in biosphere_keys2ids.items():
            result_func_map[id_] = method_config.functions[key]
        if method_config.get_defaults_from_original:
            bw_method_data = bw2data.Method(bw_method_id).load()
            bw_cf_activity_key2ids = self.activities_keys_id_map(
                [tuple[str, str](method_key) for method_key, _ in bw_method_data]
            )
            for method_key, cf in bw_method_data:
                activity_id = bw_cf_activity_key2ids[tuple(method_key)]
                result_func_map[activity_id] = lambda v, cf_=cf: v * cf_
        return result_func_map

    @staticmethod
    def get_config_schemas() -> dict:
        # todo: currently ignoring callable

        class MyGenerateJsonSchema(GenerateJsonSchema):
            def handle_invalid_for_json_schema(
                self, schema: core_schema.CoreSchema, error_info: str
            ) -> JsonSchemaValue:
                if schema["type"] == "callable":
                    logger.warning("Ignoring callable during schema generation...")
                    raise PydanticOmit
                return super().handle_invalid_for_json_schema(schema, error_info)

        adapter_schema = BWAdapterConfig.model_json_schema(
            schema_generator=MyGenerateJsonSchema
        )
        return {
            "adapter": adapter_schema,
            "activity": BrightwayActivityConfig.model_json_schema(),
            "method": BWMethodDefinition.model_json_schema(),
        }

    def activities_keys_id_map(
        self, keys: list[tuple[str, str]]
    ) -> ReversibleRemappableDictionary:
        codes = [code for _, code in keys]
        biosphere_activities = list(
            ActivityDataset.select().where(ActivityDataset.code.in_(codes))
        )
        # noinspection PyUnresolvedReferences
        return ReversibleRemappableDictionary(
            {(a.database, a.code): a.id for a in biosphere_activities}
        )



from pint import Quantity

from enbios import ureg
from enbios.models.experiment_base_models import NodeOutput


def compact_all_to(quantities: list[Quantity], use_min: bool = True) -> list[Quantity]:
    """
    Convert all quantities to the same unit, and return the compacted values
    :param quantities:
    :param use_min: use the unit of the smallest quantity (otherwise use the largest)
    :return: a list of compacted quantities with the same unit
    """
    base_func = min if use_min else max
    base_value = base_func([q.to_compact() for q in quantities])
    return [q.to(base_value.units) for q in quantities]


def get_output_in_unit(output: NodeOutput, target_unit: str) -> float:
    """
    Convert the output to a magnitude the given unit
    :param output:
    :param target_unit:
    :return:
    """
    return (
        (ureg.parse_expression(output.unit) * output.magnitude).to(target_unit).magnitude
    )



from abc import ABC, abstractmethod
from typing import Any, Optional

from enbios.base.scenario import Scenario
from enbios.generic.enbios2_logging import get_logger
from enbios.models.experiment_base_models import (
    NodeOutput,
    AdapterModel,
)
from enbios.models.experiment_models import ResultValue


class EnbiosAdapter(ABC):
    def __init__(self):
        self._config = None

    @abstractmethod
    def validate_definition(self, definition: AdapterModel):
        pass

    @abstractmethod
    def validate_config(self, config: Optional[dict[str, Any]]):
        """
        This is the first validator to be called. Validate the config. The creator may store anything in the
        adapter object here.
        :param config:
        """
        pass

    @abstractmethod
    def validate_methods(self, methods: Optional[dict[str, Any]]) -> list[str]:
        pass

    @abstractmethod
    def validate_node(self, node_name: str, node_config: Any):
        pass

    @abstractmethod
    def validate_node_output(self, node_name: str, target_output: NodeOutput) -> float:
        pass

    @abstractmethod
    def get_node_output_unit(self, node_name: str) -> str:
        pass

    @abstractmethod
    def get_method_unit(self, method_name: str) -> str:
        pass

    @abstractmethod
    def get_default_output_value(self, node_name: str) -> float:
        pass

    @abstractmethod
    def run_scenario(self, scenario: Scenario) -> dict[str, dict[str, ResultValue]]:
        """
        Run a specific scenario. The adapter should return a dictionary of the form:
            {
                node_name: {
                    method_name: ResultValue (unit, magnitude)
                }
            }
        :param scenario:
        :return:
        """
        pass

    def get_logger(self):
        return get_logger(f"ADAPTER::({self.name()})")

    @staticmethod
    @abstractmethod
    def node_indicator() -> str:
        pass

    @staticmethod
    @abstractmethod
    def get_config_schemas() -> dict[str, dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass



from bw2calc import LCA
from bw2data.backends import ActivityDataset

"""
Go through your activity dataset and assign location tuples to data["enb_locations"]
activities should have a location tuple, with increasing granularity of locations. e.g. ("EU", "ES","cat")
location levels should be consistent. meaning, since ES is on index 1, 
there should be no activity with location ("ES", "ara") (where ES is at index 0)
"""


def regionalization(lca: LCA, location_key: str = "enb_location") -> dict[str, float]:
    """
    :param lca: bw LCA object
    :param location_key:
    :return: dictionary, location > impact score
    """
    if not hasattr(lca, "inventory"):
        print("calculating inventory")
        lca.lci()

    if not hasattr(lca, "characterization_matrix"):
        lca.load_lcia_data()

    # final location -> id
    base_loc_map = {}
    # all other indices to last locs
    loc_tree = []
    for a in ActivityDataset.select(ActivityDataset).where(
        ActivityDataset.type == "process"
    ):
        # if a.type == "process":
        loc = a.data.get(location_key)
        if not isinstance(loc, tuple):
            continue
        final_loc = loc[-1]
        base_loc_map.setdefault(final_loc, []).append(a.id)
        for idx, rest in enumerate(loc[:-1]):
            if len(loc_tree) <= idx:
                loc_tree.append({})
            loc_tree[idx].setdefault(rest, set()).add(loc[idx + 1])

    loc_tree.reverse()

    res_map = {}
    # do matrix multiplication for each final location
    for loc, idxs in base_loc_map.items():
        res_map[loc] = (
            lca.characterization_matrix
            * lca.inventory[:, [lca.dicts.activity[c] for c in idxs]]
        ).sum()

    # sum up location results, per level...
    for lvl in loc_tree:
        for loc, sub_loc in lvl.items():
            # print(loc, contained)
            res_map[loc] = 0
            for cloc in sub_loc:
                res_map[loc] += res_map[cloc]

    return res_map



from pathlib import Path

import bw2data
import bw2io


def import_ecoinvent(project_name: str, dataset_path: str, database_name: str):
    project_created = False

    db_path: Path = Path(dataset_path)
    if not db_path.exists() and not db_path.is_dir():
        raise Exception(f"{db_path} does not exist or is not a directory")
    if project_name in bw2data.projects:
        raise Exception(
            f"project {project_name} already exists and "
            f"'require_fresh' is set to True."
        )
    try:
        bw2data.projects.create_project(project_name)
        project_created = True
        bw2data.projects.set_current(project_name)
        bw2io.bw2setup()

        imported = bw2io.SingleOutputEcospold2Importer(dataset_path, database_name)
        imported.apply_strategies()
        if imported.all_linked:
            imported.write_database()
    except Exception as e:
        print(e)
        if project_created:
            pass


if __name__ == "__main__":
    import_ecoinvent(
        "ei391",
        "/home/ra/projects/enbios/data/ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/datasets",
        "ei391_cutoff_ecoSpold02",
    )



from collections import Counter
from logging import getLogger
from typing import TYPE_CHECKING, Optional

from enbios.base.adapters_aggregators.adapter import EnbiosAdapter
from enbios.base.adapters_aggregators.aggregator import EnbiosAggregator
from enbios.base.adapters_aggregators.builtin import BUILTIN_AGGREGATORS
from enbios.base.adapters_aggregators.loader import load_adapter, load_aggregator
from enbios.base.scenario import Scenario
from enbios.models.environment_model import Settings
from enbios.models.experiment_base_models import (
    ExperimentConfig,
    ExperimentScenarioData,
    NodeOutput,
    AdapterModel,
    AggregationModel,
)

if TYPE_CHECKING:
    from enbios.base.experiment import Experiment

logger = getLogger(__name__)


def validate_adapters(
    experiment_adapters: list[AdapterModel],
) -> tuple[dict[str, EnbiosAdapter], list[str]]:
    """
    Validate the adapters in this experiment data

    :return: adapter-dict and method names
    """
    adapters = []
    methods = []
    for adapter_def in experiment_adapters:
        adapter = load_adapter(adapter_def)
        adapter.validate_definition(adapter_def)
        adapter.validate_config(adapter_def.config)
        adapters.append(adapter)
        adapter_methods = adapter.validate_methods(adapter_def.methods)
        methods.extend([f"{adapter.node_indicator()}.{m}" for m in adapter_methods])

    adapter_map = {adapter.name(): adapter for adapter in adapters}
    return adapter_map, methods


def validate_aggregators(
    experiment_aggregators: list[AggregationModel],
) -> dict[str, EnbiosAggregator]:
    """
    Validate the aggregators in this experiment data

    :return: a aggregator-name-Aggregator dict
    """
    aggregators = []
    for aggregator_def in experiment_aggregators:
        aggregator = load_aggregator(aggregator_def)
        aggregator.validate_config(aggregator_def.config)
        aggregators.append(aggregator)

    aggregator_names = [a.name() for a in aggregators]
    for builtin_name, aggregator in BUILTIN_AGGREGATORS.items():
        if builtin_name not in aggregator_names:
            aggregator = aggregator()

            aggregators.append(aggregator)

    return {aggregator.name(): aggregator for aggregator in aggregators}


def validate_scenarios(
    experiment_scenarios: Optional[list[ExperimentScenarioData]],
    default_scenario_name: str,
    experiment: "Experiment",
) -> list[Scenario]:
    scenarios: list[Scenario] = []

    # undefined scenarios. just one default scenario
    if not experiment_scenarios:
        experiment_scenarios = [ExperimentScenarioData(name=default_scenario_name)]

    # set names if not given
    for index, scenario_data in enumerate(experiment_scenarios):
        scenario_data.name_factory(index)

    # check for name duplicates
    name_count = Counter([s.name for s in experiment_scenarios])
    # get the scenarios that have the same name
    duplicate_names = [name for name, count in name_count.items() if count > 1]
    if duplicate_names:
        raise ValueError(f"Scenarios with the same name: {duplicate_names}")

    for index, scenario_data in enumerate(experiment_scenarios):
        scenario = validate_scenario(scenario_data, experiment)
        scenarios.append(scenario)
        scenario.prepare_tree()
    return scenarios


def validate_scenario(
    scenario_data: ExperimentScenarioData, experiment: "Experiment"
) -> Scenario:
    """
    Validate one scenario
    :param scenario_data:
    :param experiment:
    :return:
    """

    def validate_nodes(scenario_: ExperimentScenarioData) -> dict[str, float]:
        nodes = scenario_.nodes or {}
        result: dict[str, float] = {}

        for node_name_, node_output in nodes.items():
            node_ = experiment.get_structural_node(node_name_)
            adapter = experiment.get_node_adapter(node_)

            if isinstance(node_output, dict):
                node_output = NodeOutput(**node_output)
            result[node_name_] = adapter.validate_node_output(node_name_, node_output)
        return result

    scenario_nodes_outputs: dict[str, float] = validate_nodes(scenario_data)
    defined_nodes = list(scenario_nodes_outputs.keys())

    # fill up the missing activities with default values
    if not scenario_data.config.exclude_defaults:
        for node_name in experiment.structural_nodes_names:
            if node_name not in defined_nodes:
                node = experiment.get_structural_node(node_name)
                scenario_nodes_outputs[node_name] = experiment.get_node_adapter(
                    node
                ).get_default_output_value(node.name)

    return Scenario(
        experiment=experiment,  # type: ignore
        name=scenario_data.name,
        structural_nodes_outputs=scenario_nodes_outputs,
        config=scenario_data.config,
        result_tree=experiment.base_result_tree.copy(),
    )


def validate_run_scenario_setting(
    env_settings: Settings, experiment_config: ExperimentConfig, scenario_names: list[str]
):
    """
    Validate a run environmental variable that is setting the scenario
    """
    if env_settings.RUN_SCENARIOS:
        if experiment_config.run_scenarios:
            logger.info(
                "Environment variable 'RUN_SCENARIOS' is set "
                "and overwriting experiment config."
            )
        experiment_config.run_scenarios = env_settings.RUN_SCENARIOS
    if experiment_config.run_scenarios:
        for scenario in experiment_config.run_scenarios:
            if scenario not in scenario_names:
                raise ValueError(
                    f"Scenario '{scenario}' not found in experiment scenarios. "
                    f"Scenarios are: {scenario_names}"
                )



from typing import Any, TYPE_CHECKING, Callable

from enbios.models.experiment_base_models import (
    ExperimentHierarchyNodeData,
    HierarchyNodeReference,
)

if TYPE_CHECKING:
    from enbios.base.experiment import Experiment

from enbios.generic.enbios2_logging import get_logger
from enbios.generic.tree.basic_tree import BasicTreeNode
from enbios.models.experiment_models import TechTreeNodeData, ScenarioResultNodeData

logger = get_logger(__name__)


def validate_experiment_hierarchy(
    hierarchy: ExperimentHierarchyNodeData,
) -> BasicTreeNode[TechTreeNodeData]:
    # todo allow no output only when there are scenarios...
    tech_tree: BasicTreeNode[TechTreeNodeData] = BasicTreeNode.from_dict(
        hierarchy.model_dump(), dataclass=TechTreeNodeData
    )

    def validate_node_data(node: BasicTreeNode[TechTreeNodeData]) -> Any:
        good_leaf = node.is_leaf and node.data.adapter
        good_internal = not node.is_leaf and node.data.aggregator
        assert good_leaf or good_internal, (
            f"Node should have the leaf properties (id, adapter) "
            f"or non-leaf properties (children, aggregator): "
            f"{node.location_names()})"
        )
        return True

    tech_tree.recursive_apply(validate_node_data, depth_first=True)  # type: ignore
    return tech_tree


def validate_experiment_reference_hierarchy(
    hierarchy: HierarchyNodeReference,
    original_experiment_hierarchy: BasicTreeNode[TechTreeNodeData],
    get_node_aggregator_fcn: Callable,
) -> BasicTreeNode[TechTreeNodeData]:
    tech_tree: BasicTreeNode[TechTreeNodeData] = BasicTreeNode.from_dict(
        hierarchy.model_dump(), dataclass=TechTreeNodeData
    )

    def validate_node_data(node: BasicTreeNode[TechTreeNodeData]) -> Any:
        if node.is_leaf:
            if node.data.adapter:
                logger.warning(
                    "Adapter is not allowed in reference hierarchy and will be ignored"
                )
            orig_node = original_experiment_hierarchy.find_subnode_by_name(node.name)
            if not orig_node:
                raise ValueError(f"Node {node.name} not found in original hierarchy")
            node.set_data(orig_node.data)
        else:
            if not node.data.aggregator:
                orig_node = original_experiment_hierarchy.find_subnode_by_name(node.name)
                if not orig_node:
                    get_node_aggregator_fcn(node).validate_node(
                        node.name, node.data.config
                    )
                    # raise ValueError(f"Node '{node.name}' not found in original hierarchy")
                node.set_data(orig_node.data)
        return True

    tech_tree.recursive_apply(validate_node_data, depth_first=True)  # type: ignore
    return tech_tree


def recursive_resolve_outputs(
    node: BasicTreeNode[ScenarioResultNodeData], experiment: "Experiment", **kwargs
):
    # todo, does this takes default values when an node is not defined
    #  in the scenario?
    if node.is_leaf:
        return
    cancel_parts_of: set = kwargs["cancel_parents_of"]
    if any(child.id in cancel_parts_of for child in node.children):
        cancel_parts_of.add(node.id)

    aggregator = experiment.get_node_aggregator(node)
    node_output = aggregator.aggregate_node_output(node, kwargs.get("scenario_name"))
    if not node_output:
        cancel_parts_of.add(node.id)
    else:
        node.data.output = node_output



from typing import Any, Optional

from pydantic import BaseModel, model_validator

from enbios.base.adapters_aggregators.adapter import EnbiosAdapter
from enbios.base.scenario import Scenario
from enbios.generic.unit_util import get_output_in_unit
from enbios.models.experiment_base_models import AdapterModel, NodeOutput
from enbios.models.experiment_models import ResultValue


class SimpleAssignment(BaseModel):
    node_name: str
    output_unit: str
    default_output: NodeOutput
    default_impacts: Optional[dict[str, ResultValue]] = None
    scenario_outputs: Optional[dict[str, NodeOutput]] = None
    scenario_impacts: Optional[dict[str, dict[str, ResultValue]]] = None

    @model_validator(mode="before")
    @classmethod
    def validator(cls, data: Any) -> Any:
        if "default_output" not in data:
            data["default_output"] = {"unit": data["output_unit"], "magnitude": 1}
        if "default_impacts" not in data and "scenario_impacts" not in data:
            raise ValueError("Either default_impacts or scenario_impacts must be defined")
        return data


class SimpleAssignmentDefinition(BaseModel):
    methods: dict[str, str]  # name: unit
    # activities: dict[str, SimpleAssignment]


class SimpleAssignmentAdapter(EnbiosAdapter):
    @staticmethod
    def name():
        return "simple-assignment-adapter"

    def __init__(self):
        super().__init__()
        self.nodes: dict[str, SimpleAssignment] = {}  # name: node
        self.methods: Optional[dict[str, str]] = None  # name: unit

    def validate_definition(self, definition: AdapterModel):
        SimpleAssignmentDefinition(**definition.model_dump())

    def validate_config(self, config: Optional[dict[str, Any]]):
        pass

    def validate_methods(self, methods: Optional[dict[str, Any]]) -> list[str]:
        self.methods = methods
        return list(methods.keys())

    def validate_node_output(self, node_name: str, target_output: NodeOutput) -> float:
        return get_output_in_unit(target_output, self.nodes[node_name].output_unit)

    def validate_node(self, node_name: str, node_config: Any):
        self.nodes[node_name] = SimpleAssignment(
            **{**{"node_name": node_name} | node_config}
        )

    def get_node_output_unit(self, node_name: str) -> str:
        return self.nodes[node_name].output_unit

    def get_method_unit(self, method_name: str) -> str:
        return self.methods[method_name]

    def get_default_output_value(self, node_name: str) -> float:
        return self.nodes[node_name].default_output.magnitude

    # def run(self):
    #     pass

    def run_scenario(self, scenario: Scenario) -> dict[str, dict[str, ResultValue]]:
        result = {}
        for node, config in self.nodes.items():
            node_results = result.setdefault(node, {})
            for method in self.methods:
                if config.scenario_impacts:
                    if scenario.name in config.scenario_impacts:
                        node_results[method] = config.scenario_impacts[scenario.name][
                            method
                        ]
                elif config.default_impacts:
                    if method in config.default_impacts:
                        node_results[method] = config.default_impacts[method]
        return result

    @staticmethod
    def node_indicator() -> str:
        return "assign"

    @staticmethod
    def get_config_schemas() -> dict:
        return {"node_name": SimpleAssignment.model_json_schema()}



from dataclasses import dataclass
from typing import Any, Optional, Callable, Sequence

import bw2data
from bw2data.backends import Activity
from pydantic import BaseModel, ConfigDict, Field, model_validator, RootModel

from enbios.generic.enbios2_logging import get_logger
from enbios.models.experiment_base_models import NodeOutput

logger = get_logger(__name__)


class ExperimentMethodPrepData(BaseModel):
    id: tuple[str, ...]
    bw_method_unit: str


class RegionalizationConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True, strict=True)
    run_regionalization: bool = Field(False)
    select_regions: set = Field(None, description="regions to store the results for")
    set_node_regions: dict[str, Sequence[str]] = Field({}, description="Set node regions")
    clear_all_other_node_regions: Optional[bool] = Field(
        False, description="Delete all regions not in 'hierarchy' and 'set_node_regions'"
    )

    @model_validator(mode="before")
    @classmethod
    def validate(cls, data: Any) -> Any:
        if data.get("run_regionalization", False):
            if data.get("select_regions") is None:
                raise ValueError(
                    "Select regions missing for BW regionalization (field: 'select_regions')"
                )
        return data


class NonLinearMethodConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True, strict=True)
    name: str = Field(None, description="bw method tuple name")
    # bw_id: tuple[str, ...] = Field(None, description="bw method tuple name")
    functions: dict[tuple[str, str], Callable[[float], float]] = Field(None, exclude=True)
    module_path_function_name: tuple[str, str] = Field(
        None,
        description="path to a module and a function name. "
        "which holds a function that returns a "
        "'dict[tuple[str, str], Callable[[float], float]]'",
    )
    default_function: Callable[[float], float] = Field(None, init_var=False)
    get_defaults_from_original: Optional[bool] = Field(
        False,
        description="Method is already defined in BW and has characterization values. ",
    )

    @model_validator(mode="before")
    @classmethod
    def check_defaults(self, data: dict):
        has_default_function = data.get("default_function")
        has_get_defaults_from_original = data.get("get_defaults_from_original", False)
        if has_default_function and has_get_defaults_from_original:
            logger.warning(
                f"brightway nonlinear method config for method '{data['name']}' should only have one"
                f"'default_function' or 'get_defaults_from_original'. Using 'default_function'"
            )
            data["get_defaults_from_original"] = False
        if not has_default_function and not has_get_defaults_from_original:
            logger.warning(
                f"brightway nonlinear method config for method '{data['name']}' should either have"
                f"'default_function' or 'get_defaults_from_original'. Creating 'default_function'"
                f"with f(x) = x"
            )
            data["default_function"] = lambda v: v

        # either 'functions' or 'module_path_function_name' must be present
        has_functions = data.get("functions") is not None
        has_module_path_function_name = data.get("module_path_function_name")
        if not has_functions and not has_module_path_function_name:
            raise ValueError(
                "Either 'functions' or 'module_path_function_name' must be present."
            )
        return data

    # @classmethod
    # def model_json_schema(cls):
    #     # Generate the default schema
    #     schema = super().model_json_schema()
    #
    #     # Modify the schema for the 'functions' field
    #     # For example, represent it as a dictionary of strings
    #     schema['properties']['functions'] = {'type': 'object', 'additionalProperties': {'type': 'string'}}
    #     return schema


class NonLinearCharacterizationConfig(BaseModel):
    methods: dict[str, NonLinearMethodConfig] = Field(
        ...,
        description="Non linear characterization. "
        "Nested Dictionary: method_name > NonLinearMethodConfig",
    )

    @model_validator(mode="before")
    @classmethod
    def add_method_names(cls, data):
        for method_name, method_config in data["methods"].items():
            method_config["name"] = method_name
        return data


class BWAdapterConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True, strict=True)
    bw_project: str
    # methods: MethodsDataTypesExt
    use_k_bw_distributions: int = Field(
        1, description="Number of samples to use for MonteCarlo"
    )
    store_raw_results: bool = Field(
        False,
        description="If the numpy matrix of brightway should be stored in the adapter. "
        "Will be stored in `raw_results[scenario.name]`",
    )
    store_lca_object: bool = Field(
        False,
        description="If the LCA object should be stored. "
        "Will be stored in `lca_objects[scenario.name]`",
    )
    simple_regionalization: RegionalizationConfig = Field(
        description="Generate regionalized LCA", default_factory=RegionalizationConfig
    )
    nonlinear_characterization: Optional[NonLinearCharacterizationConfig] = Field(
        None, description="Nonlinear characterization"
    )


class BrightwayActivityConfig(BaseModel):
    name: str = Field(
        None, description="Search:Name of the brightway activity"
    )  # brightway name
    database: str = Field(
        None, description="Search:Name of the database to search first"
    )  # brightway database
    code: str = Field(
        None, description="Search:Brightway activity code"
    )  # brightway code
    # search and filter
    location: str = Field(None, description="Search:Location filter")  # location
    enb_location: tuple[str, ...] = Field(
        None, description="Location for regionalization"
    )
    # additional filter
    unit: str = Field(None, description="Search: unit filter of results")  # unit
    # internal-name
    default_output: NodeOutput = Field(
        None, description="Default output of the activity for all scenarios"
    )


class BWMethodModel(BaseModel):
    name: str = Field(None, description="Name for identification")
    id: tuple[str, ...] = Field(None, description="Brightway method id")


class BWMethodDefinition(RootModel):
    model_config = ConfigDict(
        title="Method definition",
        json_schema_extra={"description": "Simply a dict: name : BW method tuple"},
    )
    root: dict[str, Sequence[str]]


@dataclass
class BWActivityData:
    bw_activity: Activity
    default_output: NodeOutput


@dataclass
class BWCalculationSetup:
    name: str
    inv: list[dict[Activity, float]]
    ia: list[tuple[str, ...]]

    def register(self):
        bw2data.calculation_setups[self.name] = {"inv": self.inv, "ia": self.ia}



from typing import Callable

import numpy as np
from bw2calc.multi_lca import InventoryMatrices
from bw2data.backends import ActivityDataset

from enbios.bw2.MultiLCA_util import BaseStackedMultiLCA
from enbios.bw2.bw_models import BWCalculationSetup
from enbios.bw2.util import split_inventory


class RegioStackedMultiLCA(BaseStackedMultiLCA):
    """Wrapper class for performing LCA calculations with
    many functional units and LCIA methods.

    Needs to be passed a ``calculation_setup`` name.

    This class does not subclass the `LCA` class, and performs all
    calculations upon instantiation.

    Initialization creates `self.results`, which is a NumPy array of LCA scores,
    with rows of functional units and columns of LCIA methods.
    Ordering is the same as in the `calculation_setup`.

    """

    def __init__(
        self,
        calc_setup: BWCalculationSetup,
        select_locations: set[str],
        use_distributions: bool = False,
        location_key: str = "enb_location",
        method_activity_func_maps: dict[
            tuple[str, ...], dict[int, Callable[[float], float]]
        ] = None,
    ):
        super().__init__(
            calc_setup,
            np.zeros((len(calc_setup.inv), len(calc_setup.ia), len(select_locations))),
            use_distributions,
            method_activity_func_maps,
        )

        self.locations_base_map: dict[str, list[int]] = self.resolve_loc_basemap(
            location_key
        )

        for row, func_unit in enumerate(self.func_units):
            self.prep_demand(row, func_unit)
            for col, cf_matrix in enumerate(self.method_matrices):
                self.lca.characterization_matrix = cf_matrix

                for loc_idx, loc in enumerate(select_locations):
                    activity_ids = self.locations_base_map[loc]
                    # todo, this is a bw_utils method split_inventory
                    regional_characterized_inventory = self.lcia_calculation(
                        self.non_linear_methods_flags[col],
                        split_inventory(self.lca, activity_ids),
                    )
                    self.results[row, col, loc_idx] = (
                        regional_characterized_inventory.sum()
                    )
        self.inventory = InventoryMatrices(self.lca.biosphere_matrix, self.supply_arrays)

    def resolve_loc_basemap(self, location_key: str = "enb_location"):
        # final location -> id
        base_loc_map = {}
        # all other indices to last locs
        loc_tree = []
        for a in ActivityDataset.select(ActivityDataset).where(
            ActivityDataset.type == "process"
        ):
            # if a.type == "process":
            loc = a.data.get(location_key)
            if not isinstance(loc, tuple) and not isinstance(loc, list):
                continue
            final_loc = loc[-1]
            base_loc_map.setdefault(final_loc, []).append(a.id)
            # make tree list at least as long as length
            for idx, rest in enumerate(loc[:-1]):
                if len(loc_tree) <= idx:
                    loc_tree.append({})
                # set location default and add location
                loc_tree[idx].setdefault(rest, set()).add(loc[idx + 1])
        loc_tree.reverse()
        for level in loc_tree:
            for loc, sub_locs in level.items():
                base_loc_map.setdefault(loc, [])
                for sub_loc in sub_locs:
                    base_loc_map[loc].extend(base_loc_map[sub_loc])

        return base_loc_map



from dataclasses import field
from typing import Optional, Any

from pydantic import BaseModel, Field, model_validator

from enbios.models.experiment_base_models import StrictInputConfig, NodeOutput


class TechTreeNodeData(BaseModel):
    adapter: Optional[str] = None
    aggregator: Optional[str] = None
    # for
    config: Optional[Any] = Field(
        None, description="The identifies (method to find) a node"
    )

    @model_validator(mode="before")
    @classmethod
    def check_model(cls, data: Any) -> Any:
        if isinstance(data, dict):
            leaf_node = "adapter" in data and "config" in data
            non_leaf_node = "aggregator" in data
            assert leaf_node or non_leaf_node, (
                "Node must be either leaf ('id', 'adapter`) " "or non-leaf ('aggregator')"
            )
        return data


class ResultValue(NodeOutput):
    model_config = StrictInputConfig
    magnitude: Optional[float] = None
    multi_magnitude: Optional[list[float]] = field(default_factory=list)


class ScenarioResultNodeData(BaseModel):
    model_config = StrictInputConfig
    output: Optional[NodeOutput] = None
    results: dict[str, ResultValue] = field(default_factory=dict)
    adapter: Optional[str] = None
    aggregator: Optional[str] = None



if __name__ == "__main__":
    import bw2data as bd
    import bw2io as bi
    from bw2io.importers import SingleOutputEcospold2Importer
    import random
    import string
    from bw2data.project import projects as bw_projects
    from bw2data import databases as bw_databases

    # print(bd.projects)
    gen_project_name = True
    project_name = ""
    while gen_project_name:
        random_letters = "".join(random.choice(string.ascii_lowercase) for i in range(6))
        project_name = f"test_{random_letters}"
        gen_project_name = project_name in bd.projects

    print(project_name)

    # print(bd.projects.report())

    # project_name = "test_project"
    bd.projects.create_project(project_name)
    bd.projects.set_current(project_name)
    bi.bw2setup()

    imported = SingleOutputEcospold2Importer(
        "/home/ra/projects/enbios/data/ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/datasets",
        "ecoi_391",
    )

    imported.apply_strategies()
    # print(type(imported))
    if imported.all_linked:
        imported.write_database()
        print("db written")

    projects = list(bd.projects)
    for project in projects:
        print(project)
        bw_projects.set_current(project.name)
        print(bw_projects.dir)
        print(bd.database.databases)
        databases = bw_databases
        print(databases)



import json
from pathlib import Path

from enbios.generic.enbios2_logging import get_logger, get_module_name
from enbios.models.experiment_base_models import ExperimentData

logger = get_logger(get_module_name(__file__))


def create_experiment_schema():
    schema_file_path = Path(__file__).parent.parent.parent / "data/schema/experiment.schema.gen.json"
    schema_file_path.write_text(
        json.dumps(ExperimentData.model_json_schema(), indent=2, ensure_ascii=False),
        encoding="utf-8")
    logger.info(f"Created schema file {schema_file_path}")


if __name__ == "__main__":
    create_experiment_schema()



import csv
import math
import time
from datetime import timedelta
from pathlib import Path
from tempfile import gettempdir
from typing import Any, Optional, Union, Type, cast

from python_mermaid.diagram import MermaidDiagram
from python_mermaid.link import Link
from python_mermaid.node import Node

from enbios.base.adapters_aggregators.adapter import EnbiosAdapter
from enbios.base.adapters_aggregators.aggregator import EnbiosAggregator
from enbios.base.adapters_aggregators.builtin import BUILTIN_ADAPTERS, BUILTIN_AGGREGATORS
from enbios.base.experiment_io import resolve_input_files
from enbios.base.pydantic_experiment_validation import validate_experiment_data
from enbios.base.scenario import Scenario
from enbios.base.tree_operations import validate_experiment_hierarchy
from enbios.base.validation import (
    validate_run_scenario_setting,
    validate_scenarios,
    validate_scenario,
    validate_adapters,
    validate_aggregators,
)
from enbios.bw2.stacked_MultiLCA import StackedMultiLCA
from enbios.generic.enbios2_logging import get_logger
from enbios.generic.files import PathLike, ReadPath
from enbios.generic.tree.basic_tree import BasicTreeNode
from enbios.models.environment_model import Settings
from enbios.models.experiment_base_models import (
    ExperimentConfig,
    ExperimentScenarioData,
    NodeOutput,
    ExperimentDataResolved,
)
from enbios.models.experiment_models import (
    ScenarioResultNodeData,
    TechTreeNodeData,
)

logger = get_logger(__name__)


# Define a TypeVar that is bound to EnbiosAdapter
# EnbiosAdapterType = TypeVar("EnbiosAdapterType", bound=EnbiosAdapter)
# EnbiosAggregatorType = TypeVar("EnbiosAggregatorType", bound=EnbiosAggregator)


class Experiment:
    DEFAULT_SCENARIO_NAME = "default scenario"

    def __init__(self, data_input: Optional[Union[dict, str]] = None):
        """
        Initialize the experiment
        :param data_input: dictionary or filename to load the data from
        """
        self.env_settings = Settings()
        if not data_input:
            data_input = self.env_settings.CONFIG_FILE
            if not isinstance(data_input, str):
                raise ValueError(
                    "Experiment config-file-path must be specified as environment "
                    "variable: 'CONFIG_FILE'"
                )
        data: dict = {}
        if isinstance(data_input, str):
            config_file_path = ReadPath(data_input)
            data = config_file_path.read_data()
        else:
            data = data_input
        raw_data = validate_experiment_data(data)
        # resolve hierarchy and scenarios filepaths if present
        self.resolved_raw_data: ExperimentDataResolved = resolve_input_files(raw_data)

        # load and validate adapters and aggregators
        adapters, modules = validate_adapters(raw_data.adapters)
        self._adapters: dict[str, EnbiosAdapter] = adapters
        self.methods: list[str] = modules
        self._aggregators: dict[str, EnbiosAggregator] = validate_aggregators(
            raw_data.aggregators
        )

        # validate overall hierarchy
        self.hierarchy_root: BasicTreeNode[TechTreeNodeData] = (
            validate_experiment_hierarchy(self.resolved_raw_data.hierarchy)
        )
        self._structural_nodes: dict[str, BasicTreeNode[TechTreeNodeData]] = {}
        # validate individual nodes based on their adapter/aggregator
        for node in self.hierarchy_root.iter_all_nodes():
            if node.is_leaf:
                self.get_node_adapter(node).validate_node(node.name, node.data.config)
                self._structural_nodes[node.name] = node
            else:
                self.get_node_aggregator(node).validate_node(node.name, node.data.config)

        def recursive_convert(
            node_: BasicTreeNode[TechTreeNodeData],
        ) -> BasicTreeNode[ScenarioResultNodeData]:
            output: Optional[NodeOutput] = None
            if node_.is_leaf:
                output = NodeOutput(
                    unit=self.get_node_adapter(node_).get_node_output_unit(node_.name),
                    magnitude=0,
                )
            return BasicTreeNode(
                name=node_.name,
                data=ScenarioResultNodeData(
                    output=output,
                    adapter=node_.data.adapter,
                    aggregator=node_.data.aggregator,
                ),
                children=[recursive_convert(child) for child in node_.children],
            )

        self.base_result_tree: BasicTreeNode[ScenarioResultNodeData] = recursive_convert(
            self.hierarchy_root
        )

        self.scenarios: list[Scenario] = validate_scenarios(
            self.resolved_raw_data.scenarios, Experiment.DEFAULT_SCENARIO_NAME, self
        )
        validate_run_scenario_setting(self.env_settings, self.config, self.scenario_names)

        self._lca: Optional[StackedMultiLCA] = None
        self._execution_time: float = float("NaN")

    def get_structural_node(self, name: str) -> BasicTreeNode[TechTreeNodeData]:
        """
        Get a node by either its name
        as it is defined in the experiment data
        :param name:
        :return: BasicTreeNode[TechTreeNodeData]
        """
        node = self._structural_nodes.get(name, None)
        if not node:
            raise ValueError(f"Node with name '{name}' not found")
        return node

    def get_node_adapter(self, node: BasicTreeNode[TechTreeNodeData]) -> EnbiosAdapter:
        """
        Get the adapter of a node in the experiment hierarchy

        :param node:
        :return:
        """
        assert node.data.adapter
        return cast(
            EnbiosAdapter,
            self._get_module_by_name_or_node_indicator(
                node.data.adapter, EnbiosAdapter, node.name
            ),
        )

    def get_adapter_by_name(self, name: str) -> EnbiosAdapter:
        """
        Get an adapter by its name
        :param name:
        :return:
        """
        return cast(
            EnbiosAdapter, self._get_module_by_name_or_node_indicator(name, EnbiosAdapter)
        )

    def get_node_aggregator(
        self,
        node: Union[
            BasicTreeNode[ScenarioResultNodeData], BasicTreeNode[TechTreeNodeData]
        ],
    ) -> EnbiosAggregator:
        """
        Get the aggregator of a node
        :param node:
        :return:
        """
        assert node.data.aggregator
        return cast(
            EnbiosAggregator,
            self._get_module_by_name_or_node_indicator(
                node.data.aggregator, EnbiosAggregator, node.name
            ),
        )

    def _get_module_by_name_or_node_indicator(
        self,
        name_or_indicator: str,
        module_type: Type[Union[EnbiosAdapter, EnbiosAggregator]],
        node_name: Optional[str] = None,
    ) -> Union[EnbiosAdapter, EnbiosAggregator]:
        modules: dict[str, Union[EnbiosAdapter, EnbiosAggregator]] = (
            self._adapters if module_type == EnbiosAdapter else self._aggregators
        )
        module = modules.get(name_or_indicator)
        if module:
            return module
        # also check, if the name instead of the indicator was used
        for module in modules.values():
            if module.node_indicator() == name_or_indicator:
                return module

        node_err = f"Node '{node_name}' specifies an " if node_name else ""
        raise ValueError(
            f"{node_err} unknown {module_type.__name__}: '{name_or_indicator}'. "
            + f"Available {module_type.__name__}s are: {[m.node_indicator() for m in modules.values()]}"
        )

    def get_scenario(self, scenario_name: str) -> Scenario:
        """
        Get a scenario by its name
        :param scenario_name:
        :return:
        """
        for scenario in self.scenarios:
            if scenario.name == scenario_name:
                return scenario
        raise ValueError(f"Scenario '{scenario_name}' not found")

    def run_scenario(
        self, scenario_name: str, results_as_dict: bool = True
    ) -> Union[BasicTreeNode[ScenarioResultNodeData], dict]:
        """
        Run a specific scenario
        :param scenario_name:
        :param results_as_dict:
        :return: The result_tree converted into a dict
        """
        return self.get_scenario(scenario_name).run(results_as_dict)

    def run(
        self, results_as_dict: bool = True
    ) -> dict[str, Union[BasicTreeNode[ScenarioResultNodeData], dict]]:
        """
        Run all scenarios. Returns a dict with the scenario name as key
        and the result_tree as value
        :return: dictionary scenario-name : result_tree
        """
        # methods = [m.id for m in self.methods.values()]
        # inventories: list[list[dict[Activity, float]]] = []

        if self.config.run_scenarios:
            run_scenarios = [self.get_scenario(s) for s in self.config.run_scenarios]
            logger.info(f"Running selected scenarios: {[s.name for s in run_scenarios]}")
        else:
            run_scenarios = self.scenarios

        results = {}
        start_time = time.time()
        for scenario in run_scenarios:
            scenario.reset_execution_time()
            results[scenario.name] = scenario.run(results_as_dict)
        self._execution_time = time.time() - start_time
        return results

    @property
    def execution_time(self) -> str:
        """
        Get the execution time of the experiment (or all its scenarios) in
        a readable format
        :return: execution time in the format HH:MM:SS
        """
        if not math.isnan(self._execution_time):
            return str(timedelta(seconds=int(self._execution_time)))
        else:
            any_scenario_run = False
            scenario_results = ""
            for scenario in self.scenarios:
                if not math.isnan(scenario.get_execution_time()):
                    any_scenario_run = True
                    scenario_results += f"{scenario.name}: {scenario.execution_time}\n"
            if any_scenario_run:
                return scenario_results
            else:
                return "not run"

    def results_to_csv(
        self,
        file_path: PathLike,
        scenario_name: Optional[str] = None,
        level_names: Optional[list[str]] = None,
        include_method_units: bool = True,
    ):
        """
        Turn the results into a csv file. If no scenario name is given,
        it will export all scenarios to the same file,
        :param file_path:
        :param scenario_name: If no scenario name is given, it will
        export all scenarios to the same file,
            with an additional column for the scenario alias
        :param level_names: (list of strings) If given, the results will be
        exported with the given level names
        :param include_method_units:  (Include the units of the methods in the header)
        :return:
        """
        if scenario_name:
            scenario = self.get_scenario(scenario_name)
            scenario.results_to_csv(
                file_path,
                level_names=level_names,
                include_method_units=include_method_units,
            )
            return
        else:
            header = []
            all_rows: list = []
            for scenario in self.scenarios:
                temp_file_name = gettempdir() + f"/temp_scenario_{scenario.name}.csv"
                scenario.results_to_csv(
                    temp_file_name,
                    level_names=level_names,
                    include_method_units=include_method_units,
                )
                rows = ReadPath(temp_file_name).read_data()
                rows[0]["scenario"] = scenario.name
                if not all_rows:
                    header = list(rows[0].keys())
                    header.remove("scenario")
                    header.insert(0, "scenario")
                all_rows.extend(rows)
                if (temp_file := Path(temp_file_name)).exists():
                    temp_file.unlink()
            with Path(file_path).open("w", newline="") as csvfile:
                writer = csv.DictWriter(csvfile, header)
                writer.writeheader()
                writer.writerows(all_rows)

    def result_to_dict(self, include_output: bool = True) -> list[dict[str, Any]]:
        """
        Get the results of all scenarios as a list of dictionaries as dictionaries
        :param include_output: Include the output of each node in the tree
        :return:
        """
        return [
            scenario.result_to_dict(include_output=include_output)
            for scenario in self.scenarios
        ]

    @property
    def config(self) -> ExperimentConfig:
        """
        get the config of the experiment
        :return:
        """
        return self.resolved_raw_data.config

    def __repr__(self):
        return (
            f"Experiment: (call info() for details)\n"
            f"Structural nodes: {len(self._structural_nodes)}\n"
            f"Methods: {len(self.methods)}\n"
            f"Hierarchy (depth): {self.hierarchy_root.depth}\n"
            f"Scenarios: {len(self.scenarios)}\n"
        )

    @property
    def structural_nodes_names(self) -> list[str]:
        return list(self._structural_nodes.keys())

    @property
    def scenario_names(self) -> list[str]:
        """
        Get all scenario names
        :return: list of strings of the scenario names
        """
        return list([s.name for s in self.scenarios])

    @property
    def adapters(self) -> list[EnbiosAdapter]:
        """
        Get all adapters in a list
        :return:
        """
        return list(self._adapters.values())

    def run_scenario_config(
        self,
        scenario_config: dict,
        result_as_dict: bool = True,
        append_scenario: bool = True,
    ) -> Union[BasicTreeNode[ScenarioResultNodeData], dict]:
        """
        Run a scenario from a config dictionary. Scenario will be validated and run. An
        :param scenario_config:
        :param result_as_dict:
        :return:
        """
        scenario_data = ExperimentScenarioData(**scenario_config)
        scenario_data.name_factory(len(self.scenarios))

        extra_index = 1
        if scenario_data.name in self.scenario_names:
            scenario_data.name = scenario_data.name + f" ({extra_index})"
            extra_index += 1
        scenario = validate_scenario(scenario_data, self)
        scenario.prepare_tree()
        if append_scenario:
            self.scenarios.append(scenario)
        return scenario.run(result_as_dict)

    def info(self) -> str:
        """
        Information about the experiment
        :return:
        """
        node_rows: list[str] = []

        def print_node(node: BasicTreeNode[TechTreeNodeData], _):
            module_name: str
            if node.data.adapter:
                module_name = self.get_node_adapter(node).name()
            else:
                module_name = self.get_node_aggregator(node).name()
            node_rows.append(f"{' ' * node.level}{node.name} - {module_name}")

        self.hierarchy_root.recursive_apply(print_node, False, False, None)

        node_rows_str = "\n".join(node_rows)
        methods_str = "\n".join([f" {m}" for m in self.methods])
        return (
            f"Experiment: \n"
            f"Structural nodes: {len(self._structural_nodes)}\n"
            f"{node_rows_str}\n"
            f"Methods: {len(self.methods)}\n"
            f"{methods_str}\n"
            f"Hierarchy (depth): {self.hierarchy_root.depth}\n"
            f"Scenarios: {len(self.scenarios)}\n"
        )

    @staticmethod
    def get_module_definition(
        clazz: Union[EnbiosAdapter, EnbiosAggregator], details: bool = True
    ) -> dict[str, Any]:
        result: dict = {
            "node_indicator": clazz.node_indicator(),
        }
        if details:
            result["config"] = clazz.get_config_schemas()
        return result

    @staticmethod
    def get_builtin_adapters(details: bool = True) -> dict[str, dict[str, Any]]:
        """
        Get the built-in adapters
        :return:
        """
        result = {}
        for name, clazz in BUILTIN_ADAPTERS.items():
            result[name] = Experiment.get_module_definition(clazz, details)
        return result

    @staticmethod
    def get_builtin_aggregators(details: bool = True) -> dict[str, dict[str, Any]]:
        result = {}
        for name, clazz in BUILTIN_AGGREGATORS.items():
            result[name] = Experiment.get_module_definition(clazz, details)
        return result

    def get_all_configs(
        self, include_all_builtin_configs: bool = True
    ) -> dict[str, dict[str, dict[str, Any]]]:
        """
        Result structure:
            ```json
            {
                "adapters": { <adapter_name>: <adapter_config>},
                "aggregators": { ... }
            }
            ```
        :param include_all_builtin_configs:
        :return: all configs
        """
        result = {
            "adapters": {
                name: Experiment.get_module_definition(adapter, True)
                for name, adapter in (
                    self._adapters
                    | (BUILTIN_ADAPTERS if include_all_builtin_configs else {})
                ).items()
            },
            "aggregators": {
                name: Experiment.get_module_definition(aggregator, True)
                for name, aggregator in (
                    self._aggregators
                    | (BUILTIN_AGGREGATORS if include_all_builtin_configs else {})
                ).items()
            },
        }

        return result

    def get_method_unit(self, method: str) -> str:
        adapter_indicator, method_name = "", ""
        if "." in method:
            assert (
                method in self.methods
            ), f"Method {method} missing. Candidates: {', '.join(self.methods)}"
            adapter_indicator, method_name = method.split(".")
        else:
            assert method in self.method_names
            found = False
            for m in self.methods:
                if m.split(".")[-1] == method:
                    adapter_indicator, method_name = m.split(".")
                    found = True
                    break
            if not found:
                raise ValueError(f"Method {method} not found")
        adapter: EnbiosAdapter = cast(
            EnbiosAdapter,
            self._get_module_by_name_or_node_indicator(adapter_indicator, EnbiosAdapter),
        )
        return adapter.get_method_unit(method_name)

    @property
    def method_names(self) -> list[str]:
        return [m.split(".")[-1] for m in self.methods]

    def hierarchy2mermaid(self) -> str:
        mm_nodes_map: dict[str, Node] = {}
        nodes: list[Node] = []
        links: list[Link] = []
        for node in self.hierarchy_root.iter_all_nodes():
            mm_node = Node(node.name)
            nodes.append(mm_node)
            mm_nodes_map[node.name] = mm_node
            for child in node.children:
                links.append(Link(mm_node, mm_nodes_map[child.name]))
        return str(MermaidDiagram(nodes=nodes, links=links, orientation="top-down"))



import sys

from pydantic import ValidationError

from enbios.models.experiment_base_models import (
    ExperimentData,
    ExperimentHierarchyNodeData,
    ExperimentScenarioData,
)


def validate_experiment_data(data: dict) -> ExperimentData:
    try:
        return ExperimentData(**data)
    except ValidationError as err:
        print(err)

        try:
            ExperimentHierarchyNodeData(**data.get("hierarchy"))
            print("\nHierarchy is valid\n")
        except ValidationError as err:
            print(err)
            print(f"!!!\n!!!\nCheck: 'hierarchy': {err.errors()}\n\n")

        if scenarios := data.get("scenarios"):
            for idx, scenario in enumerate(scenarios):
                try:
                    sc = ExperimentScenarioData(**scenario)
                    print(f"Scenario {idx} ({sc.name}) is valid")
                except ValidationError as err:
                    print(err)
                    print(f"!!!\n!!!\nCheck: 'scenarios[{idx}]': {err.errors()}\n\n")

        # errors_sorted = sorted(err.errors(), key=lambda x: len(x["loc"]), reverse=True)[0]
        # print(f"!!!\n!!!\nCheck: '{errors_sorted['type']}': {errors_sorted['loc']}\n\n")
        sys.exit(1)



from typing import Optional, Union, Literal

import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler

from enbios.base.experiment import Experiment
from enbios.generic.enbios2_logging import get_logger

logger = get_logger(__name__)


class ResultsSelector:
    def __init__(
        self,
        experiment: Experiment,
        scenarios: Optional[list[str]] = None,
        methods: Optional[list[str]] = None,
    ):
        """
        Initialize the object with experiment, scenarios, and methods.

        :param experiment: The Experiment object.
        :param scenarios: A list of scenario names. If None, all scenario names from the experiment will be used.
        :param methods: A list of method names. If None, all method names from the experiment will be used.
        """
        self.experiment = experiment

        all_scenarios = [sc.name for sc in self.experiment.scenarios]
        if scenarios is not None:
            for scenario in scenarios:
                if scenario not in all_scenarios:
                    raise ValueError(f"Scenario {scenario} not found in experiment")
            self.scenarios: list[str] = scenarios
        else:
            self.scenarios: list[str] = all_scenarios

        all_methods: list[str] = self.experiment.method_names

        all_method_names: list[str] = self.experiment.method_names
        self.methods: list[str] = []
        self.method_names: list[str] = []
        if methods is not None:
            for idx, method in enumerate(methods):
                method_name = method.split(".")[-1]
                if method_name not in all_method_names:
                    raise ValueError(f"Method {method_name} not found in experiment")
                self.methods.append(method)
                self.method_names.append(method_name)
        else:
            self.methods = all_methods
            self.method_names = all_method_names

        self._complete_df = None
        self._base_df = None

    @staticmethod
    def get_result_selector(
        result_selector: Union[Experiment, "ResultsSelector"],
        scenarios: Optional[list[str]] = None,
        methods: Optional[list[str]] = None,
    ) -> "ResultsSelector":
        if isinstance(result_selector, Experiment):
            return ResultsSelector(result_selector, scenarios=scenarios, methods=methods)
        return result_selector

    def check_special_df(self, df: DataFrame):
        if not self.base_df.shape == df.shape:
            raise ValueError(
                f"Dataframe shape {df.shape} does not match base dataframe shape "
                f"{self.base_df.shape}"
            )
        if not list(self.base_df.dtypes) == list(df.dtypes):
            raise ValueError(
                f"Dataframe dtypes {df.dtypes} do not match base dataframe dtypes "
                f"{self.base_df.dtypes}"
            )

    @property
    def complete_df(self) -> DataFrame:
        if self._complete_df is None:
            data = [
                {"scenario": scenario}
                | {
                    k: v.magnitude
                    for k, v in self.experiment.get_scenario(
                        scenario
                    ).result_tree.data.results.items()
                }
                for scenario in self.experiment.scenario_names
            ]

            self._complete_df = DataFrame(data)
        return self._complete_df

    @property
    def base_df(self) -> DataFrame:
        if self._base_df is None:
            data = [
                {"scenario": scenario}
                | {
                    k: v.magnitude
                    for k, v in self.experiment.get_scenario(
                        scenario
                    ).result_tree.data.results.items()
                    if k in self.method_names
                }
                for scenario in self.scenarios
            ]

            self._base_df = DataFrame(data)
        return self._base_df

    def normalized_df(self, normalize_with_all_scenarios: bool = True) -> DataFrame:
        used_df = self.complete_df if normalize_with_all_scenarios else self.base_df
        columns = used_df.columns[1:]
        normalized_df = DataFrame()
        normalized_df["scenario"] = used_df["scenario"]
        for column in columns:
            normalized_df[column] = MinMaxScaler().fit_transform(
                used_df[column].to_numpy().reshape(-1, 1)
            )

        if normalize_with_all_scenarios:
            # delete all rows where scenario is not in self.scenarios
            normalized_df = normalized_df[normalized_df["scenario"].isin(self.scenarios)]
        return normalized_df

    def method_label_names(self, include_unit: bool = True) -> list[str]:
        return [
            (
                method
                + ("\n" + self.experiment.get_method_unit(method) if include_unit else "")
            )
            for method in self.methods
        ]

    def compare_to_baseline(self, baseline_data: ndarray):
        assert len(baseline_data) == len(self.method_names), (
            f"Baseline data must have the same length as the number of methods "
            f"({len(self.method_names)})"
        )
        # Create a copy of the original dataframe, without modifying it in place
        baseline_df = self.base_df.copy()
        # Use loc to select the columns related to methods
        # and divide them by baseline_data
        baseline_df[self.method_names] = self.base_df[self.method_names].div(
            baseline_data, axis=1
        )
        # Set the dtypes back to their original types
        original_dtypes = self.base_df[self.method_names].dtypes
        for col, dtype in original_dtypes.items():
            # print(col, dtype)
            baseline_df[col] = baseline_df[col].astype(dtype)
        return baseline_df.reset_index(drop=True)

    def validate_node_selection(self, level: int, nodes: list[str]) -> list[str]:
        if level >= self.experiment.hierarchy_root.depth:
            logger.warning(
                f"Level {level} is higher or equal (>=) than the depth of the hierarchy "
                f"({self.experiment.hierarchy_root.depth}). "
                f"Limiting to {self.experiment.hierarchy_root.depth - 1}"
            )
            level = self.experiment.hierarchy_root.depth - 1

        if nodes:
            for node_name in nodes:
                node = self.experiment.hierarchy_root.find_subnode_by_name(node_name)
                if not node:
                    raise ValueError(f"Alias {node_name} not found in hierarchy")
        else:
            nodes = [
                n.name
                for n in self.experiment.hierarchy_root.collect_all_nodes_at_level(level)
            ]
        return nodes

    def collect_tech_results(
        self,
        nodes: list[str],
        value_name: Literal["magnitude", "multi_magnitude"] = "magnitude",
    ):
        df = DataFrame()
        for scenario in self.scenarios:
            scenario_results = self.experiment.get_scenario(scenario).result_tree
            for node_name in nodes:
                node = scenario_results.find_subnode_by_name(node_name)
                assert node is not None
                # add a row
                df = pd.concat(
                    [
                        df,
                        DataFrame(
                            [
                                {
                                    "scenario": scenario,
                                    "tech": node_name,
                                }
                                | {
                                    method: getattr(value, value_name)
                                    for method, value in node.data.results.items()
                                    if method in self.method_names
                                }
                            ]
                        ),
                    ],
                    ignore_index=True,
                    copy=False,
                )
        return df



import json
from pathlib import Path
from typing import Optional, Any, Union

from pydantic import BaseModel, Field
from pyparsing import (
    ParserElement,
    Word,
    alphas,
    Or,
    Literal,
    Keyword,
    one_of,
    DelimitedList,
    Opt,
    srange,
    Group,
    alphanums,
    ParseException,
    ParseResults,
    White,
)

from enbios.generic.files import PathLike


def _create_init_def() -> ParserElement:
    init: ParserElement = Keyword("graph") + Opt(
        White() + one_of("TB TD BT RL LR").set_results_name("direction")
    )
    init = init.leave_whitespace()
    return init


def _create_node_def(group_name: Optional[str] = None) -> ParserElement:
    # Define the base components of a node
    node_name = Word(alphas, alphanums).set_results_name("name")
    attribute_expression = Group(
        Word(alphas).set_results_name("key", list_all_matches=True)
        + Literal(":")
        + Word(srange("[a-zA-Z_0-9]")).set_results_name("value", list_all_matches=True)
    ).set_results_name("config")

    @attribute_expression.set_parse_action
    def resolve_identifier(results: ParseResults):
        # turns it from key, value lists to a dictionary {<key>:<value>}
        conf = results.config
        return dict(zip(list(conf.key), list(conf.value)))

    node_shapes: dict[str, tuple[str, str]] = {
        "normal": ("[", "]"),
        "round-edge": ("(", ")"),
        "stadium-shape": ("([", "])"),
        "subroutine-shape": ("[[", "]]"),
        "cylindrical": ("[(", ")]"),
        "circle": ("((", "))"),
        "label-shape": (">", "]"),
        "rhombus": ("{", ")"),
        "hexagon": ("{{", ")}"),
        "parallelogram": ("[/", "/]"),
        "parallelogram-alt": ("[\\", "\\]"),
        "trapezoid": ("[/", "\\]"),
        "trapezoid-alt": ("[\\", "/]"),
        "double-circle": ("(((", ")))"),
    }
    # Define all possible shapes for a node
    all_expressions = [node_name]
    for shape in node_shapes.values():
        node_display_name = Word(alphas)  # Assuming this definition is correct
        all_expressions.append(
            node_name + Literal(shape[0]) + node_display_name + Literal(shape[1])
        )

    # Combine the base node definition with optional process type and unit expression
    node_def_base = Or(all_expressions)
    ada_agg_expression = Word(alphas).set_results_name(
        "module_name"
    )  # wrapping this in Opt would allow :::,key:code

    node_def = Or(
        [
            node_def_base,
            # todo: should parse the whole thing. otherwise attributes at the like
            #  this method:non-linear result in method:non
            node_def_base
            + Literal(":::")
            + ada_agg_expression
            + Opt(Literal(",") + DelimitedList(attribute_expression)),
        ]
    )

    if group_name:
        return Group(node_def).set_results_name(group_name)
    return node_def


def _create_link_def() -> ParserElement:
    simple_link_strings = ["-->"]
    simple_arrow = Or([Literal(s) for s in simple_link_strings])
    # pyparsing_common.real.set_results_name("module_name")
    text_arrow = (
        Literal("--") + Word(alphas).set_results_name("module_name") + Literal("-->")
    )
    arrow = Or((simple_arrow, text_arrow))
    return _create_node_def("l") + arrow + _create_node_def("r")


def _parse_mermaid_lines(lines: list[str]) -> list[dict]:
    init_def = _create_init_def()
    node_def = _create_node_def()
    link_def = _create_link_def()
    after_init = False
    results = []
    for line in lines:
        line = line.strip()
        if not line:
            continue
        if not after_init:
            init_res = init_def.parse_string(line)
            results.append(init_res.as_dict())
            after_init = True
            continue
        combined = Or(
            (
                Group(node_def).set_results_name("node"),
                Group(link_def).set_results_name("link"),
            )
        )
        try:
            res = combined.parse_string(line)
            results.append(res.as_dict())
        except ParseException as err:
            print(f"PARSER ERROR ON LINE: {line}")
            print(err)
        # print(line)
        # print(res.as_dict())

        after_init = True
    return results


class _Node(BaseModel):
    name: str
    module_name: Optional[str] = None
    config: Optional[dict[str, Any]] = None
    parent: Optional["_Node"] = None
    children: list["_Node"] = Field(default_factory=list)

    def __repr__(self):
        return self.name

    def set_def(self, node: "_Node"):
        self.module_name = node.module_name
        self.config = node.config


class _Link(BaseModel):
    left_node: _Node = Field(validation_alias="l")
    right_node: _Node = Field(validation_alias="r")
    module_name: Optional[str] = Field(
        None, description="The module (adaptor/aggregator) of the lower (left side) node"
    )
    # value: float = 1


def _create_tree(results: list[dict], strict_node_def: bool = True) -> _Node:
    all_nodes: list[_Node] = []

    # [x] check all nodes have only one def
    # [x] check all nodes are somehow connected (one parent check)
    # [x] check hierarchy is ok (one parent check)
    # [x] check all node set its module, config (param)

    def add_node(node_: _Node) -> _Node:
        node_names = [n.name for n in all_nodes]
        if node_.name not in node_names:
            all_nodes.append(node_)
            return node_
        else:
            node_index = node_names.index(node_.name)
            existing: _Node = all_nodes[node_index]
            if node_.module_name:
                # check: no duplicate definition
                if existing.module_name:
                    raise ValueError(f"error. just one node definition. {node_.name}")
                else:
                    existing.set_def(node_)
            return existing

    def add_nodes(link_: _Link) -> tuple[_Node, ...]:
        node_names = [n.name for n in all_nodes]
        return_nodes: list[_Node] = []
        for node_ in (link_.left_node, link_.right_node):
            node_ = add_node(node_)
            return_nodes.append(node_)
            node_names.append(node_.name)
        return tuple(return_nodes)

    for res in results:
        if "direction" in res:
            pass
        if "node" in res:
            add_node(_Node.model_validate(res["node"]))
        elif "link" in res:
            # validate data model
            link = _Link.model_validate(res["link"])
            ln, rn = add_nodes(link)
            if ln.parent:
                raise ValueError(f"Node '{ln}' can only have one parent")
            ln.parent = rn
            rn.children.append(ln)
            if link.module_name:
                ln.module_name = link.module_name

    root = None
    for node in all_nodes:
        # check all have definition
        if strict_node_def:
            if not node.module_name or not node.config:
                raise Exception(f"error. missing module_name or config. node: '{node}'")
        if not node.parent:
            if not root:
                root = node
            else:
                raise ValueError(
                    f"There can only be one root, which is '{root}. '{node}' has no parent"
                )
        # cycle check
        else:
            next_parent = node.parent
            chain = [node]
            while next_parent:
                chain.append(next_parent)
                if next_parent.name == node.name:
                    raise Exception(f"Looks like a cycle: {chain}")
                next_parent = next_parent.parent
    if not root:
        raise Exception("Tree has no root")
    return root


def _tree2enbios_hierarchy(root_node: _Node):
    def convert_node(node: _Node) -> dict:
        module_type = "aggregator" if node.children else "adapter"
        node_dict = {
            "name": node.name,
            module_type: node.module_name,
            "config": node.config,
        }
        if node.children:
            node_dict["children"] = [convert_node(kid) for kid in node.children]

        return node_dict

    return convert_node(root_node)


def _dump_hierarchy(hierarchy: dict, file_path: Path):
    json.dump(hierarchy, file_path.open("w", encoding="utf-8"), indent=4)


def convert_mermaid_file(
    file_path: PathLike, destination_path: Optional[PathLike] = None
) -> dict:
    with open(file_path, encoding="utf-8") as file:
        lines = [line for line in file]
    hierarchy = convert_mermaid_lines(lines)
    if destination_path:
        _dump_hierarchy(hierarchy, Path(destination_path))
    return hierarchy


def convert_mermaid_lines(lines: Union[str, list[str]]) -> dict:
    if isinstance(lines, str):
        lines = lines.splitlines()
    p = _parse_mermaid_lines(lines)
    tree = _create_tree(p, False)
    hierarchy = _tree2enbios_hierarchy(tree)
    return hierarchy



from typing import Callable

import numpy as np

from enbios.bw2.MultiLCA_util import BaseStackedMultiLCA
from enbios.bw2.bw_models import BWCalculationSetup


class StackedMultiLCA(BaseStackedMultiLCA):
    """Wrapper class for performing LCA calculations with
    many functional units and LCIA methods.

    Needs to be passed a ``calculation_setup`` name.

    This class does not subclass the `LCA` class, and performs all
    calculations upon instantiation.

    Initialization creates `self.results`, which is a NumPy array of LCA scores,
    with rows of functional units and columns of LCIA methods.
    Ordering is the same as in the `calculation_setup`.

    """

    def __init__(
        self,
        calc_setup: BWCalculationSetup,
        use_distributions: bool = False,
        method_activity_func_maps: dict[
            tuple[str, ...], dict[int, Callable[[float], float]]
        ] = None,
    ):
        super().__init__(
            calc_setup,
            np.zeros((len(calc_setup.inv), len(calc_setup.ia))),
            use_distributions,
            method_activity_func_maps,
        )
        self.main_loop()



import logging
from abc import ABC
from typing import Optional, Callable, Any

import numpy as np
from bw2calc import LCA
from bw2calc.multi_lca import InventoryMatrices
from bw2data import get_activity
from bw2data.backends import Activity

from enbios.bw2.bw_models import BWCalculationSetup


class BaseStackedMultiLCA(ABC):
    def __init__(
        self,
        calc_setup: BWCalculationSetup,
        results_structure: np.ndarray,
        use_distributions: bool = False,
        method_activity_func_maps: Optional[
            dict[tuple[str, ...], dict[int, Callable[[float], float]]]
        ] = None,
    ):
        self.func_units = calc_setup.inv
        self.methods = calc_setup.ia
        self.has_nonlinear_functions: bool = method_activity_func_maps is not None
        self.all = {key: 1 for func_unit in self.func_units for key in func_unit}
        self.lca = LCA(
            demand=self.all,
            method=self.methods[0],
            use_distributions=use_distributions,
        )
        self.logger = logging.getLogger(__name__)
        self.results = results_structure
        self.lca.lci()
        self.non_linear_methods_flags: list[bool] = []
        self.method_matrices = []
        self.supply_arrays = []
        self.inventory = None

        for method in self.methods:
            self.lca.switch_method(method)
            if self.has_nonlinear_functions and method in method_activity_func_maps:
                method_characterization = method_activity_func_maps[method]
                self.method_matrices.append(method_characterization)
                self.non_linear_methods_flags.append(True)
            else:
                self.method_matrices.append(self.lca.characterization_matrix)
                self.non_linear_methods_flags.append(False)

    def main_loop(self):
        for row, func_unit in enumerate(self.func_units):
            self.prep_demand(row, func_unit)

            for col, cf_matrix in enumerate(self.method_matrices):
                # logger.debug(f"Method {col}/{len(self.method_matrices)}")
                self.lca.characterization_matrix = cf_matrix
                self.lcia_calculation(self.non_linear_methods_flags[col])
                self.results[row, col] = self.lca.score

        self.inventory = InventoryMatrices(self.lca.biosphere_matrix, self.supply_arrays)

    def prep_demand(self, row: int, func_unit: dict[Activity, float]):
        self.logger.debug(f"Demand {row}/{len(self.func_units)}")
        fu_spec, fu_demand = list(func_unit.items())[0]
        if isinstance(fu_spec, int):
            fu = {fu_spec: fu_demand}
        elif isinstance(fu_spec, Activity):
            fu = {fu[0].id: fu[1] for fu in list(func_unit.items())}
        elif isinstance(fu_spec, tuple):
            a = get_activity(fu_spec)
            fu = {a.id: fu[1] for fu in list(func_unit.items())}
        else:
            raise ValueError("Unknown functional unit type")
        self.lca.lci(fu)
        self.supply_arrays.append(self.lca.supply_array)

    def lcia_calculation(
        self, non_linear: bool = False, inventory: Optional[Any] = None
    ) -> Any:
        """The actual LCIA calculation.
        Separated from ``lcia`` to be reusable in cases where the matrices are already built, e.g. ``redo_lcia`` and Monte Carlo classes.
        """
        if inventory is None:
            inventory = self.lca.inventory
        if non_linear:
            summed_inventory = inventory.sum(1)
            # initiate cf array with 0 functions
            func_array = [lambda v: 0 for _ in range(self.lca.biosphere_matrix.shape[0])]
            # assign the proper function to the cf array
            for activity_id, matrix_idx in self.lca.dicts.biosphere.items():
                if activity_id in self.lca.characterization_matrix:
                    func_array[self.lca.dicts.biosphere[activity_id]] = (
                        self.lca.characterization_matrix[activity_id]
                    )
            result = np.array([])
            for summed_row, function in zip(summed_inventory, func_array):
                result = np.append(result, function(summed_row))
            # TODO characterized_inventory should actually be m*n (m:num bioflows, n:processes)
            # here it is (1*n)
            # but we dont calc that, cuz not using the summed_inventory would probably take much longer
            self.lca.characterized_inventory = result
        else:
            self.lca.characterized_inventory = (
                self.lca.characterization_matrix * inventory
            )
        return self.lca.characterized_inventory



from os import PathLike
from typing import Union, Optional

from flatten_dict import unflatten
from frictionless import Schema
from frictionless.fields import NumberField, StringField

from enbios.generic.files import ReadPath
from enbios.models.experiment_base_models import (
    ExperimentData,
    ExperimentHierarchyNodeData,
    ExperimentScenarioData,
    ExperimentDataResolved,
    # ExperimentMethodData,
)

activities_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="database"),
        StringField(name="code"),
        StringField(name="name"),
        StringField(name="location"),
        StringField(name="unit"),
        NumberField(name="output.value", float_number=True),
        StringField(name="output.unit"),
    ]
)

activities_unflatten_map = {
    "alias": "id.alias",
    "database": "id.database",
    "code": "id.code",
    "name": "id.name",
    "location": "id.location",
    "unit": "id.unit",
    "output.value": "output.value",
    "output.unit": "output.unit",
}

methods_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="id"),
        StringField(name="id0"),
        StringField(name="id1"),
        StringField(name="id2"),
        StringField(name="id3"),
    ]
)


def unflatten_data(data: dict, structure_map: dict):
    res = {}
    for key, value in data.items():
        if key in structure_map:
            res[structure_map[key]] = value
        else:
            res[key] = value
    print(res)
    return unflatten(res, splitter="dot")


def get_abs_path(
    path: Union[str, PathLike], base_dir: Optional[Union[str, PathLike]] = None
) -> ReadPath:
    if base_dir:
        return ReadPath(base_dir) / path
    else:
        return ReadPath(path)


def resolve_input_files(raw_input: ExperimentData) -> ExperimentDataResolved:
    # hierarchy
    if isinstance(raw_input.hierarchy, str) or isinstance(raw_input.hierarchy, PathLike):
        hierarchy_file: ReadPath = get_abs_path(
            raw_input.hierarchy, raw_input.config.base_directory
        )
        if hierarchy_file.suffix == ".json":
            hierarchy_data = hierarchy_file.read_data()
        # elif hierarchy_file.suffix == ".csv":
        #     hierarchy_data = csv_tree2dict(hierarchy_file, False)
        else:
            raise Exception(f"Invalid hierarchy file: {raw_input.hierarchy}")

        raw_input.hierarchy = ExperimentHierarchyNodeData(**hierarchy_data)

    # scenarios
    if isinstance(raw_input.scenarios, str) or isinstance(raw_input.scenarios, PathLike):
        scenario_file: ReadPath = get_abs_path(
            raw_input.scenarios, raw_input.config.base_directory
        )
        if scenario_file.suffix == ".json":
            scenario_data = scenario_file.read_data()
            raw_input.scenarios = [
                ExperimentScenarioData(**scenario) for scenario in scenario_data
            ]
        else:
            raise Exception(f"Invalid scenario file: {raw_input.scenarios}")

    return ExperimentDataResolved.model_validate(raw_input.model_dump())



import concurrent.futures
import json
import math
import time
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Optional, Union, TYPE_CHECKING, Any, Callable

from enbios.base.tree_operations import validate_experiment_reference_hierarchy
from enbios.generic.enbios2_logging import get_logger
from enbios.generic.files import PathLike
from enbios.models.experiment_base_models import (
    HierarchyNodeReference,
    ScenarioConfig,
    NodeOutput,
)

# for type hinting
if TYPE_CHECKING:
    from enbios.base.experiment import Experiment
from enbios.generic.tree.basic_tree import BasicTreeNode
from enbios.models.experiment_models import (
    ScenarioResultNodeData,
    ResultValue,
    TechTreeNodeData,
)

logger = get_logger(__name__)


@dataclass
class Scenario:
    experiment: "Experiment"
    name: str
    result_tree: BasicTreeNode[ScenarioResultNodeData]

    _has_run: bool = False
    # this should be a simpler type - just str: float
    structural_nodes_outputs: dict[str, float] = field(default_factory=dict)
    # methods: Optional[dict[str, ExperimentMethodPrepData]] = None
    _execution_time: float = float("NaN")
    config: ScenarioConfig = field(default_factory=ScenarioConfig)

    def prepare_tree(self):
        """Prepare the result tree for calculating scenario outputs.
        This populates the result tree with ScenarioResultNodeData objects
        for each node, which store the output magnitude and units.
        If config is set, it also stores the BW node dict with the node.
        """

        structural_nodes_names = list(self.structural_nodes_outputs.keys())
        for result_index, node_name in enumerate(structural_nodes_names):
            try:
                structural_result_node = self.result_tree.find_subnode_by_name(node_name)
            except StopIteration:
                raise ValueError(f"Node {node_name} not found in result tree")
            structural_node = self.experiment.get_structural_node(node_name)
            structural_result_node.data.output = NodeOutput(
                unit=self.experiment.get_node_adapter(
                    structural_node
                ).get_node_output_unit(node_name),
                magnitude=self.structural_nodes_outputs[node_name],
            )
            # todo adapter/aggregator specific additional data
            # if self.experiment.config.include_bw_activity_in_nodes:
            #     node_node.data.bw_activity = bw_activity

        if self.config.exclude_defaults:

            def remove_empty_nodes(
                node: BasicTreeNode[ScenarioResultNodeData], cancel_parents_of: set[str]
            ):
                # aggregators without children are not needed
                if node.is_leaf and node.data.aggregator:
                    node.remove_self()

            for leave in self.result_tree.iter_leaves():
                if leave.name not in structural_nodes_names:
                    leave.remove_self()
            self.result_tree.recursive_apply(
                remove_empty_nodes, depth_first=True, cancel_parents_of=set()
            )

        from enbios.base.tree_operations import recursive_resolve_outputs

        self.result_tree.recursive_apply(
            recursive_resolve_outputs,
            experiment=self.experiment,
            depth_first=True,
            scenario_name=self.name,
            cancel_parents_of=set(),
        )

    @staticmethod
    def _propagate_results_upwards(
        node: BasicTreeNode[ScenarioResultNodeData], experiment: "Experiment"
    ):
        if node.is_leaf:
            return
        else:
            node.data.results = experiment.get_node_aggregator(
                node
            ).aggregate_node_result(node)

    def run(
        self, results_as_dict: bool = True
    ) -> Union[BasicTreeNode[ScenarioResultNodeData], dict]:
        # if not self._get_methods():
        #     raise ValueError(f"Scenario '{self.name}' has no methods")
        logger.info(f"Running scenario '{self.name}'")
        # distributions_config = self.experiment.config.use_k_bw_distributions
        # distribution_results = distributions_config > 1
        start_time = time.time()

        if self.experiment.config.run_adapters_concurrently:
            # Create a ThreadPoolExecutor
            with concurrent.futures.ThreadPoolExecutor() as executor:
                # Use a list comprehension to start a thread for each adapter
                futures = [
                    executor.submit(adapter.run_scenario, self)
                    for adapter in self.experiment.adapters
                ]
                # As each future completes, set the results
                for future in concurrent.futures.as_completed(futures):
                    result_data = future.result()
                    self.set_results(result_data)
        else:
            for adapter in self.experiment.adapters:
                # run in parallel:
                result_data = adapter.run_scenario(self)
                self.set_results(result_data)

        self.result_tree.recursive_apply(
            Scenario._propagate_results_upwards,
            experiment=self.experiment,
            depth_first=True,
        )

        self._has_run = True
        self._execution_time = time.time() - start_time
        return self.result_to_dict() if results_as_dict else self.result_tree

    @property
    def execution_time(self) -> str:
        if not math.isnan(self._execution_time):
            return str(timedelta(seconds=int(self._execution_time)))
        else:
            return "not run"

    def reset_execution_time(self):
        self._execution_time = float("NaN")

    def set_results(self, result_data: dict[str, Any]):
        for node_name, node_result in result_data.items():
            node = self.result_tree.find_subnode_by_name(node_name)
            if not node:
                if self.config.exclude_defaults:
                    logger.warning(
                        f"Structural node '{node_name}' not found in result tree. "
                        f"Make sure that the adapter for does not generate results for default"
                        f"nodes that are not in the scenario."
                    )
                else:
                    logger.error(f"Node '{node_name}' not found in result tree")
                continue
            node.data.results = node_result

    @staticmethod
    def wrapper_data_serializer(
        include_output: bool = True, expand_results: bool = False
    ) -> Callable[[ScenarioResultNodeData], dict]:
        def _expand_results(results: dict[str, ResultValue]) -> dict:
            """
            brings all results to the data level (one down) which is useful for csv
            :param results:
            :return:
            """
            expanded_results = {}
            for method_name, result_value in results.items():
                expanded_results[f"{method_name}_magnitude ({result_value.unit})"] = (
                    result_value.magnitude
                )
                if result_value.multi_magnitude:
                    for idx, magnitude in enumerate(result_value.multi_magnitude):
                        expanded_results[f"{method_name}_{result_value.unit}_{idx}"] = (
                            magnitude
                        )
            return expanded_results

        def data_serializer(data: ScenarioResultNodeData) -> dict:
            if expand_results:
                result: dict[str, Any] = _expand_results(data.results)
            else:
                result: dict[str, Any] = {
                    "results": {
                        method_name: result_value.model_dump(exclude_defaults=True)
                        for method_name, result_value in data.results.items()
                    }
                }
            # there might be no output, when the units dont match
            if include_output and data.output:
                if expand_results:
                    result["output_unit"] = data.output.unit
                    result["output_magnitude"] = data.output.magnitude
                else:
                    result["output"] = data.output.model_dump()
            # todo: adapter specific additional data
            # if data.bw_activity:
            #     result["bw_activity"] = data.bw_activity["code"]

            return result

        return data_serializer

    def results_to_csv(
        self,
        file_path: PathLike,
        level_names: Optional[list[str]] = None,
        include_method_units: bool = True,
        warn_no_results: bool = True,
        alternative_hierarchy: Optional[dict] = None,
    ):
        """
        Save the results (as tree) to a csv file
         :param warn_no_results:
         :param file_path:  path to save the results to
         :param level_names: names of the levels to include in the csv
         (must not match length of levels)
         :param alternative_hierarchy: An alternative hierarchy to use for the results,
          which comes from Scenario.rearrange_results.
         :param include_method_units:
        """
        if not self.result_tree:
            raise ValueError(f"Scenario '{self.name}' has no results")

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.name}' has not been run yet")

        use_tree = self.result_tree
        if alternative_hierarchy:
            use_tree = self._rearrange_results(alternative_hierarchy)

        use_tree.to_csv(
            file_path,
            include_data=True,
            level_names=level_names,
            data_serializer=self.wrapper_data_serializer(include_method_units, True),
        )

    def result_to_dict(
        self,
        include_output: bool = True,
        warn_no_results: bool = True,
        alternative_hierarchy: dict = None,
    ) -> dict[str, Any]:
        """
        Return the results as a dictionary
        :param include_output:
        :param warn_no_results:
        :param alternative_hierarchy: An alternative hierarchy to use for the results,
        which comes from Scenario.rearrange_results.
        :return:
        """

        def recursive_transform(node: BasicTreeNode[ScenarioResultNodeData]) -> dict:
            result: dict[str, Any] = {
                "name": node.name,
                **Scenario.wrapper_data_serializer(include_output)(node.data),
            }
            if node.children:
                result["children"] = [
                    recursive_transform(child) for child in node.children
                ]
            return result

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.name}' has not been run yet")
        if alternative_hierarchy:
            h = self._rearrange_results(alternative_hierarchy.copy())
            return recursive_transform(h)
        else:
            return recursive_transform(self.result_tree.copy())

    def _rearrange_results(
        self, hierarchy: dict
    ) -> BasicTreeNode[ScenarioResultNodeData]:
        hierarchy_obj = HierarchyNodeReference(**hierarchy)

        hierarchy_root: BasicTreeNode[TechTreeNodeData] = (
            validate_experiment_reference_hierarchy(
                hierarchy_obj,
                self.experiment.hierarchy_root,
                self.experiment.get_node_aggregator,
            )
        )

        def recursive_convert(
            node_: BasicTreeNode[TechTreeNodeData],
        ) -> BasicTreeNode[ScenarioResultNodeData]:
            output: Optional[NodeOutput] = None
            results = {}
            if node_.is_leaf:
                calc_data = self.result_tree.find_subnode_by_name(node_.name).data
                output = calc_data.output
                results = calc_data.results
            return BasicTreeNode(
                name=node_.name,
                data=ScenarioResultNodeData(
                    output=output,
                    results=results,
                    adapter=node_.data.adapter,
                    aggregator=node_.data.aggregator,
                ),
                children=[recursive_convert(child) for child in node_.children],
            )

        result_tree: BasicTreeNode[ScenarioResultNodeData] = recursive_convert(
            hierarchy_root
        )

        from enbios.base.tree_operations import recursive_resolve_outputs

        result_tree.recursive_apply(
            recursive_resolve_outputs,
            experiment=self.experiment,
            depth_first=True,
            cancel_parents_of=set(),
        )

        result_tree.recursive_apply(
            Scenario._propagate_results_upwards,
            experiment=self.experiment,
            depth_first=True,
        )

        return result_tree

    def get_execution_time(self) -> float:
        return self._execution_time

    def __repr__(self):
        return f"<Scenario '{self.name}'>"

    def describe(self):
        output = f"Scenario '{self.name}'\n"
        output += json.dumps(self.structural_nodes_outputs, indent=2)
        # todo: the tree instead...
        return output



from pathlib import Path
from typing import Generator, Iterator, Literal

import bw2data
import bw2io
from bw2calc import LCA
from bw2data import databases as bw_databases
from bw2data.backends import Activity, ExchangeDataset, ActivityDataset
from bw2data.project import projects as bw_projects
from bw2io import SingleOutputEcospold2Importer
from scipy.sparse import csr_matrix


def info_exchanges(act: Activity) -> dict:
    """
    Show exchanges of an activity
    :param act:
    :return:
    """
    exchanges = act.exchanges()
    for exc in exchanges:
        print(exc)
    for exc in act.upstream():
        print(exc)
    return {exc.input: exc for exc in act.exchanges()}


def iter_exchange_by_ids(
    ids: Iterator[int], batch_size: int = 1000
) -> Generator[ExchangeDataset, None, None]:
    """
    Iterate over exchanges by ids
    :param ids:
    :param batch_size:
    :return:
    """
    last_batch: bool = False
    while True:
        batch_ids: list[int] = []
        for i in range(batch_size):
            try:
                batch_ids.append(next(ids))
            except StopIteration:
                last_batch = True
                break
        for exc in ExchangeDataset.select().where(ExchangeDataset.id.in_(batch_ids)):
            yield exc
        if last_batch:
            break


def iter_activities_by_codes(
    codes: Iterator[str], batch_size: int = 1000
) -> Generator[ActivityDataset, None, None]:
    """
    Iterate over activities by codes
    :param codes:
    :param batch_size:
    :return:
    """
    last_batch: bool = False
    while True:
        batch_codes: list[str] = []
        for i in range(batch_size):
            try:
                batch_codes.append(next(codes))
            except StopIteration:
                last_batch = True
                break
        for act in ActivityDataset.select().where(ActivityDataset.code.in_(batch_codes)):
            yield act
        if last_batch:
            break


def get_activity(code: str) -> Activity:
    """
    Get activity by code
    :param code:
    :return:
    """
    activity_ds = ActivityDataset.get_or_none(ActivityDataset.code == code)
    if not activity_ds:
        raise ValueError(f"Activity with code '{code}' does not exist")
    return Activity(activity_ds)


def full_duplicate(activity: Activity, code=None, **kwargs) -> Activity:
    """
    Make a copy of an activity with its upstream exchanges
    (Otherwise, you cannot calculate the lca of the copy)
    :param activity: the activity to copy
    :param code: code of the new activity
    :param kwargs: other data for the copy
    :return: new activity
    """
    activity_copy = activity.copy(code, **kwargs)
    for upstream in activity.upstream():
        upstream.output.new_exchange(
            input=activity_copy, type=upstream["type"], magnitude=upstream.magnitude
        ).save()
    activity_copy.save()
    return activity_copy


def clean_delete(activity: Activity):
    """
    Delete an activity and its upstream exchanges.
    Otherwise, the system will be corrupted
    :param activity: The activity to delete
    """
    for link in activity.upstream():
        link.delete()
    activity.delete()


def report():
    current_ = bw2data.projects.current
    projects = list(bw2data.projects)
    for project in projects:
        print(project)
        bw_projects.set_current(project.name)
        databases = list(bw_databases)
        print(databases)
    bw2data.projects.set_current(current_)


def bw_unit_fix(unit_str: str):
    if unit_str == "kilowatt hour":
        return "kilowatt_hour"
    if unit_str == "unit":
        return "unspecificEcoinventUnit"
    return unit_str


def delete_db(project_name: str, db_name: str):
    if project_name not in bw2data.projects:
        raise ValueError(f"Project '{project_name}'does not exist")
    bw2data.projects.set_current(project_name)
    if db_name not in bw2data.databases:
        raise ValueError(f"Database '{db_name}'does not exist")
    print("proceeding to delete database ...")
    bw2data.Database(db_name).delete_instance()


def delete_project(project_name: str):
    if project_name == "default":
        print("Not gonna delete project 'default'")
        return
    if project_name not in bw2data.projects:
        print(f"Project '{project_name}'does not exist")
        return
    bw2data.projects.delete_project(project_name, True)


def delete_all_projects():
    resp = input("Are you sure you want to delete all projects? [y]")
    if resp != "y":
        print("cancelled")
        return
    for project in bw2data.projects:
        if project.name != "default":
            bw2data.projects.delete_project(project.name, True)
            print(f"Deleted {project.name}")
    bw2data.projects.purge_deleted_directories()


"""
Following two functions are from bw_utils...
"""


def _check_lca(
    lca: LCA,
    make_calculations: bool = True,
    inventory_name: Literal["inventory", "characterized_inventory"] = "inventory",
):
    if not hasattr(lca, "inventory"):
        if make_calculations:
            print("calculating inventory")
            lca.lci()
        else:
            raise ValueError("Must do lci first")
    if inventory_name == "inventory":
        return

    if not hasattr(lca, "characterization_matrix"):
        print("loading lcia data")
        lca.load_lcia_data()

    if not hasattr(lca, "characterized_inventory"):
        if make_calculations:
            print("calculating lcia")
            lca.lcia()
        else:
            raise ValueError("Must do lcia first")


def split_inventory(
    lca: LCA,
    technosphere_activities: list[int],
    inventory_name: Literal["inventory", "characterized_inventory"] = "inventory",
    make_calculations: bool = True,
) -> csr_matrix:
    """
    Split the results of a lcia calculation into groups. Each group is a list of activities, specified by their ids.
    Calculations of lci and lcia are performed when they are missing and `make_calculations` is set to `True`
    :param lca: bw LCA object
    :param technosphere_activities: list of (technosphere) activity-groups, activities are specified by their 'id'
    :param make_calculations: make lci and lcia calculations if the corresponding matrices are missing
    :return: a list of sparse matrices, which are characterized inventories split by the activity groups.
    score can be calculated by calling `sum()` for any matrix.
    """
    _check_lca(lca, make_calculations, inventory_name)
    inventory = getattr(lca, inventory_name)
    # do matrix multiplication for each final location
    return inventory[:, [lca.dicts.activity[c] for c in technosphere_activities]]


"""
This package has just one function for setting up a brightway project with a evoincent database
"""


def safe_setup_ecoinvent(
    project_name: str,
    ecoinvent_db_path: str,
    db_name: str,
    delete_project: bool = False,
    delete_if_unlinked: bool = True,
):
    """
    Initiate a project with a ecoinvent database
    :param project_name: new project name
    :param ecoinvent_db_path: dataset path of the ecoinvent database
    :param db_name: name of the new database containing the ecoinvent database
    :param delete_project: Delete existing project
    :param delete_if_unlinked: Delete project if linking didn't work
    :return:
    """
    if project_name in bw2data.projects:
        if delete_project:
            bw2data.projects.delete_project(project_name, True)
        else:
            raise KeyError(f"Project '{project_name}' already exists")
    db_path = Path(ecoinvent_db_path)
    spold_files_glob = db_path.glob("*.spold")
    if not db_path.exists:
        raise FileNotFoundError(f"{ecoinvent_db_path} does not exist")
    if not next(spold_files_glob):
        raise KeyError(f"There are no spold files in {ecoinvent_db_path}")
    bw2data.projects.create_project(project_name)
    bw2data.projects.set_current(project_name)
    bw2io.bw2setup()

    imported = SingleOutputEcospold2Importer(ecoinvent_db_path, db_name)
    imported.apply_strategies()
    # print(type(imported))
    if imported.all_linked:
        imported.write_database()
    else:
        if delete_if_unlinked:
            bw2data.projects.delete_project(project_name, True)
        else:
            imported.write_unlinked(f"{db_name}_unlinked")



import importlib.metadata

from pint import UnitRegistry

from enbios.base.unit_registry import ureg

version = importlib.metadata.version('enbios')


def get_enbios_ureg() -> UnitRegistry:
    return ureg



from typing import Optional, Union, Any

from pydantic import BaseModel, Field, model_validator, field_validator, ConfigDict

from enbios.generic.files import PathLike

StrictInputConfig = ConfigDict(extra="forbid", validate_assignment=True, strict=True)


class ExperimentConfig(BaseModel):
    model_config = StrictInputConfig
    warn_default_demand: bool = True  # todo: bring this back
    auto_aggregate: Optional[bool] = (
        True  # aggregate, with same indicator, as all children, if given.
    )
    run_adapters_concurrently: bool = True
    run_scenarios: Optional[list[str]] = (
        None  # list of scenario-name to run, ALSO AS ENV-VAR
    )
    # only used by ExperimentDataIO
    # base_directory when loading files (activities, methods, ...)
    base_directory: Optional[Union[str, PathLike]] = None
    # those are only used for testing
    debug_test_is_valid: bool = True
    debug_test_replace_bw_config: Union[bool, list[str]] = (
        True  # standard replacement, or bw-project, bw-database
    )
    debug_test_expected_error_code: Optional[int] = None
    debug_test_run: Optional[bool] = False
    note: Optional[str] = None


class AdapterModel(BaseModel):
    model_config = ConfigDict(extra="allow")
    module_path: Optional[PathLike] = None
    adapter_name: Optional[str] = Field(
        None,
        description="this this is to use inbuilt adapter "
        "(e.g. 'simple-assignment-adapter'",
    )
    config: dict = Field(default_factory=dict)
    methods: dict[str, Any] = Field(default_factory=dict)
    note: Optional[str] = Field(None, description="A note for this adapter")

    @model_validator(mode="before")
    @classmethod
    def module_specified(cls, data: Any) -> Any:
        # either module_path or module_class must be specified
        if not ("module_path" in data or "adapter_name" in data):
            raise ValueError("Either module_path or adapter_name must be specified")
        return data


class AggregationModel(BaseModel):
    module_path: Optional[PathLike] = None
    aggregator_name: str = Field(
        None,
        description="this this is to use inbuilt aggregator "
        "(e.g. 'simple-assignment-adapter'",
    )
    config: Optional[dict] = Field(default_factory=dict)
    note: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def module_specified(cls, data: Any) -> Any:
        # either module_path or module_class must be specified
        if not ("module_path" in data or "aggregator_name" in data):
            raise ValueError("Either module_path or aggregator_name must be specified")
        return data


class ExperimentHierarchyNodeData(BaseModel):
    name: str
    aggregator: str = Field(..., description="name or node-indicator of the aggregator")
    config: Optional[Any] = Field(
        None, description="setup data (id, outputs, ... arbitrary data"
    )
    children: Optional[
        list[Union["ExperimentHierarchyNodeData", "ExperimentActivityData"]]
    ] = None


class ExperimentActivityData(BaseModel):
    """
    This is the dataclass for the activities in the experiment.
    """

    name: str
    config: Any = Field(..., description="setup data (id, outputs, ... arbitrary data")
    adapter: str = Field(..., description="The adapter to be used")


class HierarchyNodeReference(BaseModel):
    """
    This is a reference to a node in the hierarchy.
    """

    name: str
    config: Optional[Any] = Field(
        None, description="setup data (id, outputs, ... arbitrary data"
    )
    aggregator: Optional[str] = None
    children: Optional[list["HierarchyNodeReference"]] = None

    @field_validator("children", mode="before")
    @classmethod
    def transform_simple_string_children(
        cls, v: list[str]
    ) -> list[Union["HierarchyNodeReference", str]]:
        return [
            HierarchyNodeReference(name=child) if isinstance(child, str) else child
            for child in v
        ]


class ScenarioConfig(BaseModel):
    exclude_defaults: Optional[bool] = (
        False  # will only use on activities that are specified in the scenario
    )


class NodeOutput(BaseModel):
    model_config = StrictInputConfig
    unit: str
    magnitude: float = 1.0

    @model_validator(mode="before")
    @classmethod
    def transform_value(cls, v: Any) -> "dict":
        if isinstance(v, dict):
            return v
        elif isinstance(v, tuple) or isinstance(v, list):
            return {"unit": v[0], "magnitude": v[1]}


class ExperimentScenarioData(BaseModel):
    model_config = StrictInputConfig
    name: Optional[str] = Field(None)
    nodes: dict[str, NodeOutput] = Field(
        None, description="name to output, null means default-output (check exists)"
    )
    config: Optional[ScenarioConfig] = Field(default_factory=ScenarioConfig)

    def name_factory(self, index: int):
        if not self.name:
            self.name = f"Scenario {index}"


class ExperimentData(BaseModel):
    """
    This class is used to store the data of an experiment.
    """

    model_config = StrictInputConfig

    adapters: list[AdapterModel] = Field(..., description="The adapters to be used")
    aggregators: list[AggregationModel] = Field(
        [], description="The aggregators to be used"
    )
    hierarchy: Union[ExperimentHierarchyNodeData, PathLike] = Field(
        ..., description="The activity hierarchy to be used in the experiment"
    )
    scenarios: Optional[Union[list[ExperimentScenarioData], PathLike]] = Field(
        None, description="The scenarios for this experiment"
    )
    config: ExperimentConfig = Field(
        default_factory=ExperimentConfig,
        description="The configuration of this experiment",
    )


class ExperimentDataResolved(BaseModel):
    """
    This class is used to store the data of an experiment.
    """

    model_config = StrictInputConfig

    adapters: list[AdapterModel] = Field(..., description="The adapters to be used")
    aggregators: list[AggregationModel] = Field(
        [], description="The aggregators to be used"
    )
    hierarchy: ExperimentHierarchyNodeData = Field(
        ..., description="The activity hierarchy to be used in the experiment"
    )
    scenarios: Optional[list[ExperimentScenarioData]] = Field(
        None, description="The scenarios for this experiment"
    )
    config: ExperimentConfig = Field(
        default_factory=ExperimentConfig,
        description="The configuration of this experiment",
    )
